##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""This file is meant to help dump information to files"""

import csv
import json
import logging
import os
from typing import Dict, List


LOG = logging.getLogger(__name__)


# TODO When we add more public methods we can get rid of this pylint disable
class Dumper:  # pylint: disable=R0903
    """
    The dumper class is intended to help write information to files.
    Currently, the supported dump file types are: csv or json.

    Attributes:
        file_name (str): The name of the file to write data to.
        file_type (str): The type of the file (either "csv" or "json") determined
            from the file name.

    Methods:
        write: Writes information to the specified output file based on the file type.
        _csv_write: Writes information to a CSV file.
        _json_write: Writes information to a JSON file.

    Example:
        CSV usage:
        ```python
        dumper = Dumper("populations.csv")
        # Eugene, OR has a population of 175096
        # Livermore, CA has a population of 86803
        population_data = {
            "City": ["Eugene", "Livermore"],
            "State": ["OR", "CA"],
            "Population": [175096, 86803]
        }
        dumper.write(population_data, "w")  # Output will be written to populations.csv
        ```

    Example:
        JSON usage:
        ```python
        dumper = Dumper("populations.json")
        population_data = {
            "OR": {"Eugene": 175096, "Portland": 641162},
            "CA": {"Livermore": 86803, "San Francisco": 815201}
        }
        dumper.write(population_data, "w")  # Output will be written to populations.json
        ```
    """

    def __init__(self, file_name: str):
        """
        Initializes the Dumper class and validates the file type.

        Args:
            file_name: The name of the file to write data to.

        Raises:
            ValueError: If the file type is not supported. Supported types are CSV and JSON.
        """
        supported_types = ["csv", "json"]

        valid_file = False
        for stype in supported_types:
            if file_name.endswith(stype):
                valid_file = True
                self.file_type = stype

        if not valid_file:
            raise ValueError(f"Invalid file type for {file_name}. Supported file types are: {supported_types}.")

        self.file_name = file_name

    def write(self, info_to_write: Dict, fmode: str):
        """
        Writes information to the specified output file.

        This method determines the file type and calls the appropriate
        method to write the data.

        Args:
            info_to_write: The information to write to the output file.
            fmode: The file write mode ("w" for write, "a" for append, etc.).
        """
        if self.file_type == "csv":
            self._csv_write(info_to_write, fmode)
        elif self.file_type == "json":
            self._json_write(info_to_write, fmode)

    def _csv_write(self, csv_to_dump: Dict[str, List], fmode: str):
        """
        Writes information to a CSV file.

        Args:
            csv_to_dump: The data to write to the CSV file. Keys are column
                headers and values are column values.
            fmode: The file write mode ("w" for write, "a" for append, etc.).
        """
        # If we have statuses to write, create a csv writer object and write to the csv file
        with open(self.file_name, fmode) as outfile:
            csv_writer = csv.writer(outfile)
            if fmode == "w":
                csv_writer.writerow(csv_to_dump.keys())
            csv_writer.writerows(zip(*csv_to_dump.values()))

    def _json_write(self, json_to_dump: Dict[str, Dict], fmode: str):
        """
        Writes information to a JSON file.

        Args:
            json_to_dump: The data to write to the JSON file.
            fmode: The file write mode ("w" for write, "a" for append, etc.).
        """
        # Appending to json requires file mode to be r+ for json.load
        if fmode == "a":
            fmode = "r+"

        with open(self.file_name, fmode) as outfile:
            # If we're appending, read in the existing file data
            if fmode == "r+":
                file_data = json.load(outfile)
                json_to_dump.update(file_data)
                outfile.seek(0)
            # Write to the outfile
            json.dump(json_to_dump, outfile)


def dump_handler(dump_file: str, dump_info: Dict):
    """
    Handles the process of creating a Dumper object and writing data
    to an output file.

    This function checks if the specified dump file exists to determine
    the appropriate file write mode (append or write). It then creates
    a Dumper object and writes the provided information to the file,
    logging the process.

    Args:
        dump_file: The filepath to the file where data will be dumped.
        dump_info: A dictionary containing the information to be written
            to the `dump_file`.
    """
    # Create a dumper object to help us write to dump_file
    dumper = Dumper(dump_file)

    # Get the correct file write mode and log message
    fmode = "a" if os.path.exists(dump_file) else "w"
    write_type = "Writing" if fmode == "w" else "Appending"
    LOG.info(f"{write_type} to {dump_file}...")

    # Write the output
    dumper.write(dump_info, fmode)
    LOG.info(f"{write_type} complete.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module of all Merlin-specific exception types.
"""

# Pylint complains that these exceptions are no different from Exception
# but we don't care, we just need new names for exceptions here
# pylint: disable=W0235

__all__ = (
    "RetryException",
    "SoftFailException",
    "HardFailException",
    "InvalidChainException",
    "RestartException",
    "NoWorkersException",
)


class RetryException(Exception):
    """
    Exception to signal that a step needs to be
    retried.
    """

    def __init__(self):
        super().__init__()


class SoftFailException(Exception):
    """
    Exception for non-fatal Merlin errors. Workflow
    should continue.
    """

    def __init__(self):
        super().__init__()


class HardFailException(Exception):
    """
    Exception for fatal Merlin errors. Should cause
    workflow to terminate.
    """

    def __init__(self):
        super().__init__()


class InvalidChainException(Exception):
    """
    Exception for invalid Merlin step Directed Acyclic Graphs (DAGs).
    """

    def __init__(self):
        super().__init__()


class RestartException(Exception):
    """
    Exception to signal that a step needs to call
    the restart command if present , else retry.
    """

    def __init__(self, message):
        super().__init__(message)


class NoWorkersException(Exception):
    """
    Exception to signal that no workers were started
    to process a non-empty queue(s).
    """

    def __init__(self, message):
        super().__init__(message)


class MerlinInvalidTaskServerError(Exception):
    """
    Exception to signal that an invalid task server was provided.
    """

    def __init__(self, message):
        super().__init__(message)


class BackendNotSupportedError(Exception):
    """
    Exception to signal that the provided backend is not supported by Merlin.
    """

    def __init__(self, message):
        super().__init__(message)


###############################
# Database-Related Exceptions #
###############################


class EntityNotFoundError(Exception):
    """
    Fallback error for entities that can't be found.
    """


class StudyNotFoundError(EntityNotFoundError):
    """
    Exception to signal that the study you were looking for cannot be found in
    Merlin's database.
    """


class RunNotFoundError(EntityNotFoundError):
    """
    Exception to signal that the run you were looking for cannot be found in
    Merlin's database.
    """


class WorkerNotFoundError(EntityNotFoundError):
    """
    Exception to signal that the worker you were looking for cannot be found in
    Merlin's database.
    """


class UnsupportedDataModelError(Exception):
    """
    Exception to signal that the data model you're trying to use is not supported.
    """


class EntityManagerNotSupportedError(Exception):
    """
    Exception to signal that the provided entity manager is not supported by Merlin.
    """



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Merlin CLI `example` command module.

This module defines the `ExampleCommand` class, which integrates into the Merlin
command-line interface to support downloading and setting up example workflows.
"""

# pylint: disable=duplicate-code

import logging
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.examples.generator import list_examples, setup_example


LOG = logging.getLogger("merlin")


class ExampleCommand(CommandEntryPoint):
    """
    Handles `example` CLI command for downloading built-in Merlin examples.

    Methods:
        add_parser: Adds the `example` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `example` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `example` command parser will be added.
        """
        example: ArgumentParser = subparsers.add_parser(
            "example",
            help="Generate an example merlin workflow.",
            formatter_class=RawTextHelpFormatter,
        )
        example.set_defaults(func=self.process_command)
        example.add_argument(
            "workflow",
            action="store",
            type=str,
            help="The name of the example workflow to setup. Use 'merlin example list' to see available options.",
        )
        example.add_argument(
            "-p",
            "--path",
            action="store",
            type=str,
            default=None,
            help="Specify a path to write the workflow to. Defaults to current working directory",
        )

    def process_command(self, args: Namespace):
        """
        CLI command to set up or list Merlin example workflows.

        This function either lists all available example workflows or sets
        up a specified example workflow to be run in the root directory. The
        behavior is determined by the `workflow` argument.

        Args:
            args: Parsed command-line arguments, which may include:\n
                - `workflow`: The action to perform; should be "list"
                to display all examples or the name of a specific example
                workflow to set up.
                - `path`: The directory where the example workflow
                should be set up. Only applicable when `workflow` is not "list".
        """
        if args.workflow == "list":
            print(list_examples())
        else:
            print(banner_small)
            setup_example(args.workflow, args.path)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Main CLI parser setup for the Merlin command-line interface.

This module defines the primary argument parser for the `merlin` CLI tool,
including custom error handling and integration of all available subcommands.
"""

import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from merlin import VERSION
from merlin.ascii_art import banner_small
from merlin.cli.commands import ALL_COMMANDS


DEFAULT_LOG_LEVEL = "INFO"


class HelpParser(ArgumentParser):
    """
    This class overrides the error message of the argument parser to
    print the help message when an error happens.

    Methods:
        error: Override the error message of the `ArgumentParser` class.
    """

    def error(self, message: str):
        """
        Override the error message of the `ArgumentParser` class.

        Args:
            message: The error message to log.
        """
        sys.stderr.write(f"error: {message}\n")
        self.print_help()
        sys.exit(2)


def build_main_parser() -> ArgumentParser:
    """
    Set up the command-line argument parser for the Merlin package.

    Returns:
        An `ArgumentParser` object with every parser defined in Merlin's codebase.
    """
    parser = HelpParser(
        prog="merlin",
        description=banner_small,
        formatter_class=RawDescriptionHelpFormatter,
        epilog="See merlin <command> --help for more info",
    )
    parser.add_argument("-v", "--version", action="version", version=VERSION)
    parser.add_argument(
        "-lvl",
        "--level",
        type=str,
        default=DEFAULT_LOG_LEVEL,
        help="Set log level: DEBUG, INFO, WARNING, ERROR [Default: %(default)s]",
    )
    subparsers = parser.add_subparsers(dest="subparsers", required=True)

    for command in ALL_COMMANDS:
        command.add_parser(subparsers)

    return parser



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `examples` package provides resources for learning and setting up Merlin workflows.

Modules:
    examples.py: Contains example specification files for Merlin workflows, along with
        detailed explanations of each block in the specification.
    generator.py: Provides utilities for managing and generating example workflows.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
`StudyManager` module for managing study entities in the Merlin database.

This module defines a manager class responsible for the creation, retrieval,
and deletion of studies. It also ensures appropriate cleanup of associated
run entities when a study is deleted.
"""

from __future__ import annotations

from typing import List

from merlin.db_scripts.data_models import StudyModel
from merlin.db_scripts.entities.study_entity import StudyEntity
from merlin.db_scripts.entity_managers.entity_manager import EntityManager


# Purposefully ignoring this pylint message as each entity will have different parameter requirements
# pylint: disable=arguments-differ,arguments-renamed


class StudyManager(EntityManager[StudyEntity, StudyModel]):
    """
    Manager class for handling study entities.

    The `StudyManager` interacts with the underlying storage backend to create,
    retrieve, and delete study records. It also supports cleanup of related
    run entities during deletion to ensure data integrity.

    Attributes:
        backend (backends.results_backend.ResultsBackend): Backend interface used to persist
            and query study data.
        db (db_scripts.merlin_db.MerlinDatabase): Reference to the full `MerlinDatabase`,
            used for cross-entity operations (e.g., deleting associated runs).

    Methods:
        create: Create a new study if it doesn't already exist.
        get: Retrieve a study by ID or name.
        get_all: Retrieve all study entities.
        delete: Delete a study, with optional cleanup of related runs.
        delete_all: Delete all studies and optionally their runs.
        set_db_reference: Set reference to the MerlinDatabase for cross-entity access.
    """

    def create(self, study_name: str) -> StudyEntity:
        """
        Create a study if it does not already exist.

        If a study with the given name is not found in the database,
        a new study entity is created and persisted.

        Args:
            study_name (str): The name of the study to create.

        Returns:
            The newly created or existing study entity.
        """
        return self._create_entity_if_not_exists(
            entity_class=StudyEntity,
            model_class=StudyModel,
            identifier=study_name,
            log_message_exists=f"Study with name '{study_name}' already has an entry in the database.",
            log_message_create=f"Study with name '{study_name}' does not yet have an entry in the database. Creating one.",
            name=study_name,
        )

    def get(self, study_id_or_name: str) -> StudyEntity:
        """
        Retrieve a study by its ID or name.

        Args:
            study_id_or_name (str): The unique study ID or name to look up.

        Returns:
            The corresponding study entity.

        Raises:
            StudyNotFoundError: If no study matches the given ID or name.
        """
        return self._get_entity(StudyEntity, study_id_or_name)

    def get_all(self) -> List[StudyEntity]:
        """
        Retrieve all study entities stored in the database.

        Returns:
            A list of all available study entities.
        """
        return self._get_all_entities(StudyEntity, "study")

    def delete(self, study_id_or_name: str, remove_associated_runs: bool = True):
        """
        Delete a study and optionally its associated runs.

        If `remove_associated_runs` is True, all runs linked to the study
        will be deleted before the study itself is removed.

        Args:
            study_id_or_name (str): The ID or name of the study to delete.
            remove_associated_runs (bool, optional): Whether to delete runs
                associated with the study. Defaults to True.
        """

        def cleanup_study(study):
            if remove_associated_runs:
                for run_id in study.get_runs():
                    self.db.runs.delete(run_id)

        self._delete_entity(StudyEntity, study_id_or_name, cleanup_fn=cleanup_study)

    def delete_all(self, remove_associated_runs: bool = True):
        """
        Delete all studies in the database, and optionally their associated runs.

        Args:
            remove_associated_runs (bool, optional): Whether to delete runs
                linked to each study. Defaults to True.
        """
        self._delete_all_by_type(
            get_all_fn=self.get_all,
            delete_fn=self.delete,
            entity_name="studies",
            remove_associated_runs=remove_associated_runs,
        )

    def set_db_reference(self, db: MerlinDatabase):  # noqa: F821  pylint: disable=undefined-variable
        """
        Set a reference to the main Merlin database object for cross-entity operations.

        This allows the manager to access other entity managers (e.g., runs) when
        performing operations like cleanup during deletions.

        Args:
            db (db_scripts.merlin_db.MerlinDatabase): The database object that provides
                access to related entity managers.
        """
        self.db = db



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `mixins` package provides reusable mixin classes that encapsulate common behaviors
shared across multiple entity types in the Merlin system.

These mixins are designed to be composed into larger classes (e.g., entity models or
entity managers) without enforcing inheritance from a shared base, offering lightweight
extensions to class behavior.

Modules:
    name.py: Defines the [`NameMixin`][db_scripts.entities.mixins.name.NameMixin], which provides name-based
        access and utility methods for entities with a `name` attribute.
    queue_management.py: Defines the
        [`QueueManagementMixin`][db_scripts.entities.mixins.queue_management.QueueManagementMixin],
        which supports retrieval and filtering of task queues for entities with a `queues` field.
    run_management.py: Defines the
        [`RunManagementMixin`][db_scripts.entities.mixins.run_management.RunManagementMixin],
        which provides functionality for managing run associations on entities that support saving,
        reloading, and tracking linked run IDs.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module defines the abstract base class `EntityManager`, which provides a generic framework
for managing database-backed entities in the Merlin system.

`EntityManager` is designed to be subclassed for specific entity types (e.g., studies, runs,
workers), allowing consistent implementation of core CRUD operations (Create, Read, Update, Delete)
across different entity managers. It provides shared helper methods for common operations such as
creating entities if they do not exist, retrieving entities by ID, and deleting entities.
"""

import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, List, Optional, Type, TypeVar

from merlin.backends.results_backend import ResultsBackend
from merlin.db_scripts.data_models import BaseDataModel
from merlin.db_scripts.entities.db_entity import DatabaseEntity
from merlin.exceptions import RunNotFoundError, StudyNotFoundError, WorkerNotFoundError


T = TypeVar("T", bound=DatabaseEntity)
M = TypeVar("M", bound=BaseDataModel)

LOG = logging.getLogger("merlin")


class EntityManager(Generic[T, M], ABC):
    """
    Abstract base class for managing database entity lifecycles.

    This class defines a common interface and helper methods for managing entities stored in the
    database, such as studies, runs, and workers. Subclasses must implement methods to create,
    retrieve, and delete these entities using the provided backend.

    Generic Parameters:
        T (DatabaseEntity): The entity class managed by this manager.
        M (BaseDataModel): The data model class corresponding to the entity.

    Attributes:
        backend: The backend interface used to persist and retrieve entity data.

    Methods:
        create: Abstract method to create a new entity.
        get: Abstract method to retrieve a single entity by its identifier.
        get_all: Abstract method to retrieve all entities of this type.
        delete: Abstract method to delete a specific entity by its identifier.
        delete_all: Abstract method to delete all entities of this type.
        _create_entity_if_not_exists: Creates an entity only if it does not already exist in the backend.
        _get_entity: Retrieves an entity instance using its identifier.
        _get_all_entities: Retrieves all entities of a specific type from the backend.
        _delete_entity: Deletes an individual entity, optionally calling a cleanup function before deletion.
        _delete_all_by_type: Deletes all entities of a certain type using the provided getter and deleter functions.
    """

    def __init__(self, backend: ResultsBackend):
        """
        Initialize the EntityManager with a backend.

        Args:
            backend: The backend interface used to persist and retrieve entities.
        """
        self.backend = backend
        self.db = None  # Subclasses can set this by creating a set_db_reference method

    @abstractmethod
    def create(self, *args: Any, **kwargs: Any) -> T:
        """
        Create a new entity.

        This method must be implemented by subclasses to define how a specific entity is created.

        Args:
            *args (Any): Optional positional arguments for create context.
            **kwargs (Any): Optional keyword arguments for create context.

        Returns:
            The newly created entity instance.

        Raises:
            NotImplementedError: If a subclass has not implemented this method.
        """
        raise NotImplementedError("Subclasses of `EntityManager` must implement a `create` method.")

    @abstractmethod
    def get(self, identifier: str) -> T:
        """
        Retrieve a single entity by its identifier.

        Args:
            identifier (str): The unique identifier of the entity.

        Returns:
            The entity instance corresponding to the identifier.

        Raises:
            NotImplementedError: If a subclass has not implemented this method.
        """
        raise NotImplementedError("Subclasses of `EntityManager` must implement a `get` method.")

    @abstractmethod
    def get_all(self) -> List[T]:
        """
        Retrieve all entities managed by this entity manager.

        Returns:
            A list of all entities of the specified type.

        Raises:
            NotImplementedError: If a subclass has not implemented this method.
        """
        raise NotImplementedError("Subclasses of `EntityManager` must implement a `get_all` method.")

    @abstractmethod
    def delete(self, identifier: str, **kwargs: Any):
        """
        Delete a specific entity by its identifier.

        Args:
            identifier (str): The unique identifier of the entity to delete.
            **kwargs (Any): Optional keyword arguments for deletion context (e.g., cleanup flags).

        Raises:
            NotImplementedError: If a subclass has not implemented this method.
        """
        raise NotImplementedError("Subclasses of `EntityManager` must implement a `delete` method.")

    @abstractmethod
    def delete_all(self, **kwargs: Any):
        """
        Delete all entities managed by this entity manager.

        Args:
            **kwargs (Any): Optional keyword arguments for deletion context.

        Raises:
            NotImplementedError: If a subclass has not implemented this method.
        """
        raise NotImplementedError("Subclasses of `EntityManager` must implement a `delete_all` method.")

    def _create_entity_if_not_exists(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        entity_class: Type[T],
        model_class: Type[M],
        identifier: str,
        log_message_exists: str,
        log_message_create: str,
        **model_kwargs: Any,
    ) -> T:
        """
        Helper method to create an entity if it does not already exist.

        Args:
            entity_class (Type[T]): The class used to load and instantiate the entity.
            model_class (Type[M]): The data model class used to create a new entity if needed.
            identifier (str): The unique identifier of the entity.
            log_message_exists (str): Log message to emit if the entity already exists.
            log_message_create (str): Log message to emit when creating a new entity.
            **model_kwargs (Any): Keyword arguments passed to the model class constructor.

        Returns:
            The existing or newly created entity.
        """
        try:
            entity = entity_class.load(identifier, self.backend)
            LOG.info(log_message_exists)
        except (WorkerNotFoundError, StudyNotFoundError, RunNotFoundError):
            LOG.info(log_message_create)
            model = model_class(**model_kwargs)
            entity = entity_class(model, self.backend)
            entity.save()
        return entity

    def _get_entity(self, entity_class: Type[T], identifier: str) -> T:
        """
        Retrieve a single entity instance using its identifier.

        Args:
            entity_class (Type[T]): The class used to load the entity.
            identifier (str): The unique identifier of the entity.

        Returns:
            The loaded entity instance.
        """
        return entity_class.load(identifier, self.backend)

    def _get_all_entities(self, entity_class: Type[T], entity_type: str) -> List[T]:
        """
        Retrieve all entities of a specific type from the backend.

        Args:
            entity_class (Type[T]): The class used to instantiate each entity.
            entity_type (str): The type identifier used by the backend to filter entities.

        Returns:
            A list of all entities of the specified type.
        """
        all_entities = self.backend.retrieve_all(entity_type)
        if not all_entities:
            return []
        return [entity_class(entity_data, self.backend) for entity_data in all_entities]

    def _delete_entity(self, entity_class: Type[T], identifier: str, cleanup_fn: Optional[Callable] = None):
        """
        Delete a single entity, optionally performing cleanup beforehand.

        Args:
            entity_class (Type[T]): The class used to load and delete the entity.
            identifier (str): The unique identifier of the entity.
            cleanup_fn (Optional[Callable]): Optional function to perform cleanup before deletion.
        """
        entity = self._get_entity(entity_class, identifier)
        if cleanup_fn:
            cleanup_fn(entity)
        entity_class.delete(identifier, self.backend)

    def _delete_all_by_type(self, get_all_fn: Callable, delete_fn: Callable, entity_name: str, **delete_kwargs: Any):
        """
        Delete all entities of a specific type using provided getter and deleter functions.

        Args:
            get_all_fn (Callable): Function to retrieve all entities.
            delete_fn (Callable): Function to delete an individual entity by ID.
            entity_name (str): Human-readable name of the entity type, used for logging.
            **delete_kwargs (Any): Additional keyword arguments passed to the delete function.
        """
        all_entities = get_all_fn()
        if all_entities:
            for entity in all_entities:
                delete_fn(entity.get_id(), **delete_kwargs)
        else:
            LOG.warning(f"No {entity_name} found in the database.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module provides functionality to handle overriding variables in a spec file
via the command-line interface. It includes functions to validate and replace variables
in the spec file or environment block based on user-provided overrides.
"""

import logging
from copy import deepcopy
from typing import Dict


LOG = logging.getLogger(__name__)


def error_override_vars(override_vars: Dict[str, str], spec_filepath: str):
    """
    Warns the user if any given variable name in the override list is not found
    in the original spec file.

    Args:
        override_vars: A dictionary of variables to override, where keys are
            variable names and values are their corresponding new values. Can
            be `None` if no overrides are provided.
        spec_filepath: The file path to the original spec file.

    Raises:
        ValueError: If any variable name in `override_vars` is not found in the
            spec file.
    """
    if override_vars is None:
        return
    with open(spec_filepath, "r") as ospec:
        original_text = ospec.read()
    for variable in override_vars.keys():
        if variable not in original_text:
            raise ValueError(f"Command line override variable '{variable}' not found in spec file '{spec_filepath}'.")


def replace_override_vars(env: Dict[str, str], override_vars: Dict[str, str]) -> Dict[str, str]:
    """
    Replaces variables in the given environment block with the provided override
    values.

    Args:
        env: The environment block, represented as a dictionary where keys are
            environment variable names and values are their corresponding values.
        override_vars: A dictionary of variables to override, where keys are
            variable names and values are their corresponding new values. Can be
            `None` if no overrides are provided.

    Returns:
        A new environment block with the override variables replaced.
    """
    if override_vars is None:
        return env
    result = deepcopy(env)
    for key, val in env.items():
        for var_name, var_val in override_vars.items():
            if var_name in val:
                result[key][var_name] = var_val
    return result



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing database entities related to studies.

This module defines the `StudyEntity` class, which extends the abstract base class
[`DatabaseEntity`][db_scripts.entities.db_entity.DatabaseEntity], to encapsulate study-specific
operations and behaviors.
"""

import logging

from merlin.db_scripts.data_models import StudyModel
from merlin.db_scripts.entities.db_entity import DatabaseEntity
from merlin.db_scripts.entities.mixins.name import NameMixin
from merlin.db_scripts.entities.mixins.run_management import RunManagementMixin


LOG = logging.getLogger("merlin")


class StudyEntity(DatabaseEntity[StudyModel], RunManagementMixin, NameMixin):
    """
    A class representing a study in the database.

    This class provides methods to interact with and manage a study's data, including
    retrieving, adding, and removing run IDs from the list of runs associated with the
    study, as well as saving or deleting the study itself from the database.

    Attributes:
        entity_info (db_scripts.data_models.StudyModel): An instance of the `StudyModel`
            class containing the study's metadata.
        backend (backends.results_backend.ResultsBackend): An instance of the `ResultsBackend`
            class used to interact with the database.

    Methods:
        __repr__:
            Provide a string representation of the `StudyEntity` instance.

        __str__:
            Provide a human-readable string representation of the `StudyEntity` instance.

        reload_data:
            Reload the latest data for this study from the database.

        get_id:
            Retrieve the unique ID of the study. _Implementation found in
                [`DatabaseEntity.get_id`][db_scripts.entities.db_entity.DatabaseEntity.get_id]._

        get_additional_data:
            Retrieve any additional metadata associated with the study. _Implementation found in
                [`DatabaseEntity.get_additional_data`][db_scripts.entities.db_entity.DatabaseEntity.get_additional_data]._

        get_name:
            Retrieve the name of the study.

        get_runs:
            Retrieve the IDs of the runs associated with this study.

        add_run:
            Add a run ID to the list of runs.

        remove_run:
            Remove a run ID from the list of runs.

        save:
            Save the current state of the study to the database.

        load:
            (classmethod) Load a `StudyEntity` instance from the database by its ID or name.

        delete:
            (classmethod) Delete a study from the database by its ID or name. Optionally, remove all associated runs.
    """

    @classmethod
    def _get_entity_type(cls) -> str:
        return "study"

    def __repr__(self) -> str:
        """
        Provide a string representation of the `StudyEntity` instance.

        Returns:
            A human-readable string representation of the `StudyEntity` instance.
        """
        return (
            f"StudyEntity("
            f"id={self.get_id()}, "
            f"name={self.get_name()}, "
            f"runs={self.get_runs()}, "
            f"additional_data={self.get_additional_data()}, "
            f"backend={self.backend.get_name()})"
        )

    def __str__(self) -> str:
        """
        Provide a string representation of the `StudyEntity` instance.

        Returns:
            A human-readable string representation of the `StudyEntity` instance.
        """
        study_id = self.get_id()
        return (
            f"Study with ID {study_id}\n"
            f"------------{'-' * len(study_id)}\n"
            f"Name: {self.get_name()}\n"
            f"Runs:\n{self.construct_run_string()}"
            f"Additional Data: {self.get_additional_data()}\n\n"
        )



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for shutting down Merlin workers.

This module defines the `StopWorkersCommand` class, which handles the `stop-workers`
subcommand in the Merlin CLI. It provides functionality to stop running workers that
are connected to a task server such as Celery.
"""

# pylint: disable=duplicate-code

import logging
from argparse import ArgumentParser, Namespace

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.router import stop_workers
from merlin.spec.specification import MerlinSpec
from merlin.utils import verify_filepath


LOG = logging.getLogger("merlin")


class StopWorkersCommand(CommandEntryPoint):
    """
    Handles `stop-workers` CLI command for shutting down Merlin workers.

    Methods:
        add_parser: Adds the `stop-workers` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `stop-workers` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `stop-workers` command parser will be added.
        """
        stop: ArgumentParser = subparsers.add_parser("stop-workers", help="Attempt to stop all task server workers.")
        stop.set_defaults(func=self.process_command)
        stop.add_argument(
            "--spec",
            type=str,
            default=None,
            help="Path to a Merlin YAML spec file from which to read worker names to stop.",
        )
        stop.add_argument(
            "--task_server",
            type=str,
            default="celery",
            help="Task server type from which to stop workers.\
                                Default: %(default)s",
        )
        stop.add_argument("--queues", type=str, default=None, nargs="+", help="specific queues to stop")
        stop.add_argument(
            "--workers",
            type=str,
            action="store",
            nargs="+",
            default=None,
            help="regex match for specific workers to stop",
        )

    def process_command(self, args: Namespace):
        """
        CLI command for stopping all workers.

        This function stops any active workers connected to a user's task server.
        If the `--spec` argument is provided, this function retrieves the names of
        workers from a the spec file and then issues a command to stop them.

        Args:
            args: Parsed command-line arguments, which may include:\n
                - `spec`: Path to the specification file to load worker names.
                - `task_server`: Address of the task server to send the stop command to.
                - `queues`: List of queue names to filter the workers.
                - `workers`: List of specific worker names to stop.
        """
        print(banner_small)
        worker_names = []

        # Load in the spec if one was provided via the CLI
        if args.spec:
            spec_path = verify_filepath(args.spec)
            spec = MerlinSpec.load_specification(spec_path)
            worker_names = spec.get_worker_names()
            for worker_name in worker_names:
                if "$" in worker_name:
                    LOG.warning(f"Worker '{worker_name}' is unexpanded. Target provenance spec instead?")

        # Send stop command to router
        stop_workers(args.task_server, worker_names, args.queues, args.workers)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module defines the default values for every block in a Merlin specification file.

Merlin specification files are used to configure workflows, tasks, and resources in
Merlin. This module provides the default values for each block in the spec file, ensuring
that workflows can execute even if certain fields are omitted by the user. These defaults
serve as a fallback mechanism to maintain consistency and simplify the configuration process.

This module serves as a reference for the default configuration of Merlin specification files.
"""

DESCRIPTION = {"description": {}}

BATCH = {"batch": {"type": "local", "dry_run": False, "shell": "/bin/bash"}}

ENV = {"env": {"variables": {}, "sources": [], "labels": {}, "dependencies": {}}}

STUDY_STEP_RUN = {"task_queue": "merlin", "shell": "/bin/bash", "max_retries": 30}

PARAMETER = {"global.parameters": {}}

MERLIN = {
    "merlin": {
        "resources": {"task_server": "celery", "overlap": False, "workers": None},
        "samples": None,
    }
}

WORKER = {"steps": ["all"], "nodes": None, "batch": None}

SAMPLES = {
    "generate": {"cmd": "echo 'Insert sample-generating command here'"},
    "level_max_dirs": 25,
}

# Values of the form (step key to search for, default value if no step key found)
VLAUNCHER_VARS = {
    "MERLIN_NODES": ("nodes", 1),
    "MERLIN_PROCS": ("procs", 1),
    "MERLIN_CORES": ("cores per task", 1),
    "MERLIN_GPUS": ("gpus", 0),
}



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""This module represents everything that goes into server configuration"""

import enum
import logging
import os
import random
import string
import subprocess
from importlib import resources
from io import BufferedReader
from typing import Dict, Tuple

import yaml

from merlin.config.config_filepaths import MERLIN_HOME
from merlin.server.server_util import (
    CONTAINER_TYPES,
    MERLIN_SERVER_CONFIG,
    MERLIN_SERVER_SUBDIR,
    AppYaml,
    RedisConfig,
    RedisUsers,
    ServerConfig,
)


LOG = logging.getLogger("merlin")

# Default values for configuration
IMAGE_NAME = "redis_latest.sif"
PROCESS_FILE = "merlin_server.pf"
CONFIG_FILE = "redis.conf"
REDIS_URL = "docker://redis"
LOCAL_APP_YAML = "./app.yaml"

PASSWORD_LENGTH = 256


class ServerStatus(enum.Enum):
    """
    Represents different states that a server can be in.

    Attributes:
        RUNNING (int): Indicates the server is running and operational. Numeric value: 0.
        NOT_INITIALIZED (int): Indicates the server has not been initialized yet. Numeric value: 1.
        MISSING_CONTAINER (int): Indicates the server is missing a required container. Numeric value: 2.
        NOT_RUNNING (int): Indicates the server is not currently running. Numeric value: 3.
        ERROR (int): Indicates the server encountered an error. Numeric value: 4.
    """

    RUNNING = 0
    NOT_INITIALIZED = 1
    MISSING_CONTAINER = 2
    NOT_RUNNING = 3
    ERROR = 4


def generate_password(length, pass_command: str = None) -> str:
    """
    Generates a password for a Redis container.

    If a specific command is provided, the password will be generated using the output
    of the given command. Otherwise, a random password will be created by combining
    characters (letters, digits, and special symbols) based on the specified length.

    Args:
        length (int): The desired length of the password.
        pass_command (str, optional): A shell command to generate the password.
            If provided, the command's output will be used as the password.

    Returns:
        The generated password.
    """
    if pass_command:
        process = subprocess.run(pass_command, shell=True, capture_output=True, text=True)
        return process.stdout.strip()

    characters = list(string.ascii_letters + string.digits + "!@#$%^&*()")

    random.shuffle(characters)

    password = []
    for _ in range(length):
        password.append(random.choice(characters))

    random.shuffle(password)
    return "".join(password)


def parse_redis_output(redis_stdout: BufferedReader) -> Tuple[bool, str]:
    """
    Parses the Redis output from a Redis container.

    This function processes the Redis container's output to extract necessary information,
    such as configuration details and server state. It determines whether the server was
    successfully initialized and ready to accept connections, or if an error occurred.

    Args:
        redis_stdout (BufferedReader): A buffered reader object containing the Redis container's output.

    Returns:
        A tuple containing:\n
            - A boolean indicating whether the server was successfully initialized and ready.
            - A dictionary containing parsed configuration values if successful, or an error message otherwise.
    """
    if redis_stdout is None:
        return False, "None passed as redis output"
    server_init = False
    redis_config = {}
    line = redis_stdout.readline()
    while line != b"" and line is not None:
        if not server_init:
            values = [ln for ln in line.split() if b"=" in ln]
            for val in values:
                key, value = val.split(b"=")
                redis_config[key.decode("utf-8")] = value.strip(b",").strip(b".").decode("utf-8")
            if b"Server initialized" in line:
                server_init = True
        if b"Ready to accept connections" in line:
            return True, redis_config
        if b"aborting" in line or b"Fatal error" in line:
            return False, line.decode("utf-8")
        line = redis_stdout.readline()

    return False, "Reached end of redis output without seeing 'Ready to accept connections'"


def copy_container_command_files(config_dir: str) -> bool:
    """
    Copies YAML files containing command instructions for container types to the specified configuration directory.

    Args:
        config_dir (str): The path to the configuration directory where the YAML files will be copied.

    Returns:
        True if all files are successfully copied or already exist. False otherwise.
    """
    files = [i + ".yaml" for i in CONTAINER_TYPES]
    for file in files:
        file_path = os.path.join(config_dir, file)
        if os.path.exists(file_path):
            LOG.info(f"{file} already exists.")
            continue
        LOG.info(f"Copying file {file} to configuration directory.")
        try:
            with resources.path("merlin.server", file) as config_file:
                with open(file_path, "w") as outfile, open(config_file, "r") as infile:
                    outfile.write(infile.read())
        except OSError:
            LOG.error(f"Destination location {config_dir} is not writable.")
            return False
    return True


def create_server_config() -> bool:
    """
    Creates the main configuration file for the Merlin server in the Merlin configuration directory.

    This function checks for the existence of the Merlin configuration directory and creates a default
    server configuration if none exists. It also copies necessary container command files, applies the
    server configuration to `app.yaml`, and initializes the server configuration directory. If the
    configuration already exists, it will not overwrite it.

    Returns:
        True if the configuration is successfully created and applied. False otherwise.
    """
    # Check for ~/.merlin/ directory
    if not os.path.exists(MERLIN_HOME):
        LOG.error(f"Unable to find main merlin configuration directory at {MERLIN_HOME}")
        return False

    # Create ~/.merlin/server/ directory if it doesn't already exist
    config_dir = os.path.join(MERLIN_HOME, MERLIN_SERVER_SUBDIR)
    if not os.path.exists(config_dir):
        LOG.info("Unable to find exisiting server configuration.")
        LOG.info(f"Creating default configuration in {config_dir}")
        try:
            os.mkdir(config_dir)
        except OSError as err:
            LOG.error(err)
            return False

    # Copy container-specific yaml files to ~/.merlin/server/
    if not copy_container_command_files(config_dir):
        return False

    # Load Merlin Server Configuration and apply it to merlin_server/app.yaml
    with resources.path("merlin.server", MERLIN_SERVER_CONFIG) as merlin_server_config:
        with open(merlin_server_config) as f:  # pylint: disable=C0103
            main_server_config = yaml.load(f, yaml.Loader)

            server_config = load_server_config(main_server_config)
            if not server_config:
                LOG.error('Try to run "merlin server init" again to reinitialize values.')
                return False

            if not os.path.exists(server_config.container.get_config_dir()):
                LOG.info("Creating merlin server directory.")
                os.mkdir(server_config.container.get_config_dir())

            LOG.info("Applying merlin server configuration to ./merlin_server/app.yaml")
            server_app_yaml = AppYaml()  # By default this will point to ./merlin_server/app.yaml
            server_app_yaml.update_data(main_server_config)
            server_app_yaml.write()

    return True


def config_merlin_server():
    """
    Configures the Merlin server with necessary settings, including username and password.

    This function sets up the Merlin server by generating and storing a password file, creating a user file,
    and configuring Redis settings. If the password or user files already exist, it skips the respective
    setup steps. The function ensures that default and environment-specific users are added to the user file.
    """

    server_config = pull_server_config()
    if not server_config:
        LOG.error('Try to run "merlin server init" again to reinitialize values.')
        return False

    pass_file = server_config.container.get_pass_file_path()
    if os.path.exists(pass_file):
        LOG.info("Password file already exists. Skipping password generation step.")
    else:
        password = generate_password(PASSWORD_LENGTH)

        with open(pass_file, "w+") as f:  # pylint: disable=C0103
            f.write(password)

        LOG.info("Creating password file for merlin server container.")

    user_file = server_config.container.get_user_file_path()
    if os.path.exists(user_file):
        LOG.info("User file already exists.")
    else:
        redis_users = RedisUsers(user_file)
        redis_config = RedisConfig(server_config.container.get_config_path())
        redis_config.set_password(server_config.container.get_container_password())
        redis_users.add_user(user="default", password=server_config.container.get_container_password())
        redis_users.add_user(user=os.environ.get("USER"), password=server_config.container.get_container_password())
        redis_users.write()
        redis_config.write()

        LOG.info(f"User {os.environ.get('USER')} created in user file for merlin server container")

    return None


def load_server_config(server_config: Dict) -> ServerConfig:
    """
    Given a dictionary containing server configuration values, load them into a
    [`ServerConfig`][server.server_util.ServerConfig] instance.

    Args:
        server_config: A dictionary containing server configuration values. Should have
            a 'container' entry and a 'process' entry.

    Returns:
        A `ServerConfig` object containing the loaded server configuration.
    """
    return_data = {}
    return_data.update(server_config)
    format_needed_keys = ["command", "run_command", "stop_command", "pull_command"]
    process_needed_keys = ["status", "kill"]

    config_dir = os.path.join(MERLIN_HOME, MERLIN_SERVER_SUBDIR)

    if "container" in server_config:
        if "format" in server_config["container"]:
            format_file = os.path.join(config_dir, server_config["container"]["format"] + ".yaml")
            with open(format_file, "r") as ff:  # pylint: disable=C0103
                format_data = yaml.load(ff, yaml.Loader)
                for key in format_needed_keys:
                    if key not in format_data[server_config["container"]["format"]]:
                        LOG.error(f'Unable to find necessary "{key}" value in format config file {format_file}')
                        return None
                return_data.update(format_data)
        else:
            LOG.error('Unable to find "format" in server_config object.')
            return None
    else:
        LOG.error('Unable to find "container" in server_config object.')
        return None

    # Checking for process values that are needed for main functions and defaults
    if "process" not in server_config:
        LOG.error('Unable to find "process" in server_config object.')
        return None

    for key in process_needed_keys:
        if key not in server_config["process"]:
            LOG.error(f'Process necessary "{key}" command configuration not found in server_config object.')
            return None

    return ServerConfig(return_data)


def pull_server_config(app_yaml_path: str = None) -> ServerConfig:
    """
    Retrieves the main configuration file and its corresponding format configuration file for the Merlin server.

    This function reads the `app.yaml` configuration file and additional format-specific configuration files
    to construct a complete configuration dictionary. It validates the presence of required keys in the format
    and process configurations. If any required configuration is missing, an error is logged and `None` is returned.

    Returns:
        An instance of [`ServerConfig`][server.server_util.ServerConfig] containing all necessary configuration values.
    """
    app_yaml_file = app_yaml_path if app_yaml_path is not None else LOCAL_APP_YAML
    merlin_app_yaml = AppYaml(app_yaml_file)
    server_config = merlin_app_yaml.get_data()
    return load_server_config(server_config)


def pull_server_image() -> bool:
    """
    Fetches the server image and ensures the necessary configuration files are in place.

    This function retrieves the server image from a specified URL and saves it locally if it does not already exist.
    Additionally, it copies the default Redis configuration file to the appropriate location if it is missing.
    The function relies on the server configuration to determine the image URL, image path, and configuration file details.

    Returns:
        True if the server image and configuration file are successfully set up, False if an error occurs.
    """
    server_config = pull_server_config()
    if not server_config:
        LOG.error('Try to run "merlin server init" again to reinitialize values.')
        return False

    config_dir = server_config.container.get_config_dir()
    config_file = server_config.container.get_config_name()
    image_url = server_config.container.get_image_url()
    image_path = server_config.container.get_image_path()

    if not os.path.exists(image_path):
        LOG.info(f"Fetching redis image from {image_url}")
        subprocess.run(
            server_config.container_format.get_pull_command()
            .strip("\\")
            .format(command=server_config.container_format.get_command(), image=image_path, url=image_url)
            .split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    else:
        LOG.info(f"{image_path} already exists.")

    if not os.path.exists(os.path.join(config_dir, config_file)):
        LOG.info("Copying default redis configuration file.")
        try:
            with resources.path("merlin.server", config_file) as file:
                with open(os.path.join(config_dir, config_file), "w") as outfile, open(file, "r") as infile:
                    outfile.write(infile.read())
        except OSError as exc:
            LOG.error(f"Destination location {config_dir} is not writable. Raised from:\n{exc}")
            return False
    else:
        LOG.info("Redis configuration file already exist.")

    return True


def get_server_status() -> ServerStatus:
    """
    Determines the current status of the server.

    This function checks the server's state by verifying the existence of necessary files,
    including configuration files, the container image, and the process file. It also checks
    if the server process is actively running.

    Returns:
        An enum value representing the server's current state:\n
            - `ServerStatus.NOT_INITIALIZED`: The server has not been initialized.
            - `ServerStatus.MISSING_CONTAINER`: The server container image is missing.
            - `ServerStatus.NOT_RUNNING`: The server process is not running.
            - `ServerStatus.RUNNING`: The server is actively running.
    """
    server_config = pull_server_config()
    if not server_config:
        return ServerStatus.NOT_INITIALIZED

    if not os.path.exists(server_config.container.get_config_dir()):
        return ServerStatus.NOT_INITIALIZED

    if not os.path.exists(server_config.container.get_image_path()):
        return ServerStatus.MISSING_CONTAINER

    if not os.path.exists(server_config.container.get_pfile_path()):
        return ServerStatus.NOT_RUNNING

    pf_data = pull_process_file(server_config.container.get_pfile_path())
    parent_pid = pf_data["parent_pid"]

    check_process = subprocess.run(
        server_config.process.get_status_command().strip("\\").format(pid=parent_pid).split(),
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )

    if check_process.stdout == b"":
        return ServerStatus.NOT_RUNNING

    return ServerStatus.RUNNING


def check_process_file_format(data: Dict) -> bool:
    """
    Validates the format of a process file.

    This function checks if the given process file data (in dictionary format) contains all the
    required keys: "parent_pid", "image_pid", "port", and "hostname".

    Args:
        data (Dict): The process file data to validate.

    Returns:
        True if the process file contains all required keys, False otherwise.
    """
    required_keys = ["parent_pid", "image_pid", "port", "hostname"]
    for key in required_keys:
        if key not in data:
            return False
    return True


def pull_process_file(file_path: str) -> Dict:
    """
    Reads and parses data from a process file.

    This function attempts to load the contents of a process file located at the specified
    file path. If the file exists and its format is valid, the data is returned as a dictionary.
    If the format is invalid or the file cannot be processed, `None` is returned.

    Args:
        file_path (str): The path to the process file.

    Returns:
        A dictionary containing the data from the process file if the format is valid.
    """
    with open(file_path, "r") as f:  # pylint: disable=C0103
        data = yaml.load(f, yaml.Loader)
        if check_process_file_format(data):
            return data
    return None


def dump_process_file(data: Dict, file_path: str) -> bool:
    """
    Writes process data to a specified file.

    This function takes a dictionary containing process data and writes it to the specified
    file path in YAML format. Before writing, the function validates the format of the data.
    If the data format is invalid, the function returns `False`. If the operation is successful,
    it returns `True`.

    Args:
        data (Dict): The process data to be written to the file.
        file_path (str): The path to the file where the data will be written.

    Returns:
        True if the data is successfully written to the file, False if the data
            format is invalid or the operation fails.
    """
    if not check_process_file_format(data):
        return False
    with open(file_path, "w+") as f:  # pylint: disable=C0103
        yaml.dump(data, f, yaml.Dumper)
    return True



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module defines the `DatabaseEntity` abstract base class, which serves as a common interface
for interacting with database entities such as studies, runs, and workers. The `DatabaseEntity`
class provides a standardized structure for managing these entities, including methods for
saving, deleting, reloading, and retrieving additional metadata.

Classes that inherit from `DatabaseEntity` must implement the abstract methods defined in the base
class, ensuring consistency across different types of database entities. This abstraction reduces
code duplication and promotes maintainability by centralizing shared functionality.
"""

import logging
from abc import ABC, abstractmethod
from typing import Dict, Generic, Type, TypeVar

from merlin.backends.results_backend import ResultsBackend
from merlin.db_scripts.data_models import BaseDataModel
from merlin.exceptions import EntityNotFoundError, RunNotFoundError, StudyNotFoundError, WorkerNotFoundError


# Type variable for entity models
T = TypeVar("T", bound=BaseDataModel)
# Type variable for entity classes
E = TypeVar("E", bound="DatabaseEntity")

LOG = logging.getLogger(__name__)


class DatabaseEntity(Generic[T], ABC):
    """
    Abstract base class for database entities such as runs, studies, and workers.

    This class defines the common interface and behavior for interacting with
    database entities, including saving, deleting, and reloading data.

    Attributes:
        entity_info (T): The data model containing information about the entity.
        backend (backends.results_backend.ResultsBackend): The backend instance used
            to interact with the database.

    Methods:
        __repr__:
            Provide a string representation of the entity.

        __str__:
            Provide a human-readable string representation of the entity.

        reload_data:
            Reload the latest data for this entity from the database.

        get_id:
            Get the unique ID for this entity.

        get_additional_data:
            Get any additional data saved to this entity.

        save:
            Save the current state of this entity to the database.

        _post_save_hook:
            Hook called after saving the entity to the database. Subclasses can override
            this method to add additional behavior.

        load:
            Load an entity from the database by its ID.

        delete:
            Delete an entity from the database by its ID.

        entity_type:
            Get the type of this entity for database operations.

        _get_entity_type:
            Get the entity type for database operations (used internally).
    """

    # Mapping of entity types to their error classes
    _error_classes = {
        "logical_worker": WorkerNotFoundError,
        "physical_worker": WorkerNotFoundError,
        "run": RunNotFoundError,
        "study": StudyNotFoundError,
    }

    def __init__(self, entity_info: T, backend: ResultsBackend):
        """
        Initialize a `DatabaseEntity` instance.

        Args:
            entity_info (T): The data model containing information about the entity.
            backend (backends.results_backend.ResultsBackend): The backend instance used to
                interact with the database.
        """
        self.entity_info: T = entity_info
        self.backend: ResultsBackend = backend

    @abstractmethod
    def __repr__(self) -> str:
        """
        Provide a string representation of the entity.
        """
        raise NotImplementedError("Subclasses of `DatabaseEntity` must implement a `__repr__` method.")

    @abstractmethod
    def __str__(self) -> str:
        """
        Provide a human-readable string representation of the entity.
        """
        raise NotImplementedError("Subclasses of `DatabaseEntity` must implement a `__str__` method.")

    @property
    def entity_type(self) -> str:
        """
        Get the type of this entity for database operations.

        Returns:
            The type name as a string.
        """
        return self._get_entity_type()

    @classmethod
    def _get_entity_type(cls) -> str:
        """
        Get the entity type for database operations.

        Returns:
            The type name as a string.
        """
        # Default implementation based on class name
        # Can be overridden by subclasses if needed
        return cls.__name__.lower().replace("entity", "")

    def reload_data(self):
        """
        Reload the latest data for this entity from the database.

        Raises:
            (exceptions.EntityNotFoundError): If an entry for this entity was not found in the database.
        """
        entity_id = self.get_id()
        updated_entity_info = self.backend.retrieve(entity_id, self.entity_type)
        if not updated_entity_info:
            error_class = self._error_classes.get(self.entity_type, EntityNotFoundError)
            raise error_class(f"{self.entity_type.capitalize()} with ID {entity_id} not found in the database.")
        self.entity_info = updated_entity_info

    def get_id(self) -> str:
        """
        Get the unique ID for this entity.

        Returns:
            The unique ID for this entity.
        """
        return self.entity_info.id

    def get_additional_data(self) -> Dict:
        """
        Get any additional data saved to this entity.

        Returns:
            A dictionary of additional data.
        """
        self.reload_data()
        return self.entity_info.additional_data

    def save(self):
        """
        Save the current state of this entity to the database.
        """
        self.backend.save(self.entity_info)
        self._post_save_hook()

    def _post_save_hook(self):
        """
        Hook called after saving the entity to the database.
        Subclasses can override to add additional behavior.
        """

    @classmethod
    def load(cls: Type[E], entity_identifier: str, backend: ResultsBackend) -> E:
        """
        Load an entity from the database by its ID.

        Args:
            entity_identifier: The ID of the entity to load.
            backend (backends.results_backend.ResultsBackend): A `ResultsBackend` instance.

        Returns:
            An instance of the entity.

        Raises:
            (exceptions.EntityNotFoundError): If an entry for the entity was not found in the database.
        """
        entity_info = backend.retrieve(entity_identifier, cls._get_entity_type())
        if not entity_info:
            error_class = cls._error_classes.get(cls._get_entity_type(), EntityNotFoundError)
            raise error_class(f"{cls._get_entity_type().capitalize()} with ID {entity_identifier} not found in the database.")

        return cls(entity_info, backend)

    @classmethod
    def delete(cls, entity_identifier: str, backend: ResultsBackend):
        """
        Delete an entity from the database by its ID.

        Args:
            entity_identifier: The ID of the entity to delete.
            backend (backends.results_backend.ResultsBackend): A ResultsBackend instance.
        """
        entity_type = cls._get_entity_type()
        LOG.debug(f"Deleting {entity_type} with ID '{entity_identifier}' from the database...")
        backend.delete(entity_identifier, entity_type)
        LOG.info(f"{entity_type.capitalize()} with ID '{entity_identifier}' has been successfully deleted.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for project-wide utility functions.
"""
import getpass
import logging
import os
import pickle
import re
import socket
import subprocess
import sys
from contextlib import contextmanager
from copy import deepcopy
from datetime import datetime, timedelta
from types import SimpleNamespace
from typing import Any, Callable, Dict, Generator, List, Tuple, Union

import numpy as np
import pkg_resources
import psutil
import yaml
from tabulate import tabulate


LOG = logging.getLogger(__name__)
ARRAY_FILE_FORMATS = ".npy, .csv, .tab"
DEFAULT_FLUX_VERSION = "0.48.0"


def get_user_process_info(user: str = None, attrs: List[str] = None) -> List[Dict]:
    """
    Return a list of process information for all of the user's running processes.

    This function retrieves and returns details about the currently running processes
    for a specified user. If no user is specified, it defaults to the current user.
    It can also return information for all users if specified.

    Args:
        user: The username for which to retrieve process information. If set to
            'all_users', retrieves processes for all users. Defaults to the current
            user's username if not provided.
        attrs: A list of attributes to include in the process information. Defaults
            to ["pid", "name", "username", "cmdline"] if None. If "username" is not
            included in the list, it will be added.

    Returns:
        A list of dictionaries containing the specified attributes for each process
            belonging to the specified user or all users if 'all_users' is specified.
    """
    if attrs is None:
        attrs = ["pid", "name", "username", "cmdline"]

    if "username" not in attrs:
        attrs.extend("username")

    if user is None:
        user = getpass.getuser()

    if user == "all_users":
        return [p.info for p in psutil.process_iter(attrs=attrs)]
    return [p.info for p in psutil.process_iter(attrs=attrs) if user in p.info["username"]]


def check_pid(pid: int, user: str = None) -> bool:
    """
    Check if a given process ID (PID) is in the process list for a specified user.

    This function determines whether a specific PID is currently running
    for the specified user. If no user is specified, it defaults to the
    current user. It can also check for all users if specified.

    Args:
        pid: The process ID to check for in the process list.
        user: The username for which to check the process. If set to 'all_users',
            checks processes for all users. Defaults to the current user's username
            if not provided.

    Returns:
        True if the specified PID is found in the process list for the given user,
            False otherwise.
    """
    user_processes = get_user_process_info(user=user)
    for process in user_processes:
        if int(process["pid"]) == pid:
            return True
    return False


def get_pid(name: str, user: str = None) -> List[int]:
    """
    Return the process ID(s) (PID) of processes with the specified name.

    This function retrieves the PID(s) of all running processes that match
    the given name for a specified user. If no user is specified, it defaults
    to the current user. It can also retrieve PIDs for all users if specified.

    Args:
        name: The name of the process to search for.
        user: The username for which to retrieve the process IDs. If set to
            'all_users', retrieves processes for all users. Defaults to the
            current user's username if not provided.

    Returns:
        A list of PIDs for processes matching the specified name. Returns None
            if no matching processes are found.
    """
    user_processes = get_user_process_info(user=user)
    name_list = [p["pid"] for p in user_processes if name in p["name"]]
    if name_list:
        return name_list
    return None


def get_procs(name: str, user: str = None) -> List[Tuple[int, str]]:
    """
    Return a list of tuples containing the process ID (PID) and command line
    of processes with the specified name.

    This function retrieves all running processes that match the given name
    for a specified user. If no user is specified, it defaults to the current
    user. It can also retrieve processes for all users if specified.

    Args:
        name: The name of the process to search for.
        user: The username for which to retrieve the process information.
            If set to 'all_users', retrieves processes for all users.
            Defaults to the current user's username if not provided.

    Returns:
        A list of tuples, each containing the PID and command line of processes
            matching the specified name. Returns an empty list if no matching
            processes are found.
    """
    user_processes = get_user_process_info(user=user)
    procs = [(p["pid"], p["cmdline"]) for p in user_processes if name in p["name"]]
    return procs


def is_running_psutil(cmd: str, user: str = None) -> bool:
    """
    Determine if a process with the given command is currently running.

    This function checks for the existence of any running processes that
    match the specified command. It uses the `psutil` library to gather
    process information instead of making a call to the 'ps' command.

    Args:
        cmd: The command or command line snippet to search for in running
            processes.
        user: The username for which to check running processes. If set to
            'all_users', checks processes for all users. Defaults to the
            current user's username if not provided.

    Returns:
        True if at least one matching process is found; otherwise, False.
    """
    user_processes = get_user_process_info(user=user)
    return any(cmd in " ".join(p["cmdline"]) for p in user_processes)


def is_running(name: str, all_users: bool = False) -> bool:
    """
    Determine if a process with the specified name is currently running.

    This function checks for the existence of a running process with the
    provided name by executing the 'ps' command. It can be configured to
    check processes for all users or just the current user.

    Args:
        name: The name of the process to search for.
        all_users: If True, checks for processes across all users. Defaults
            to False, which checks only the current user's processes.

    Returns:
        True if a process with the specified name is found; otherwise, False.
    """
    cmd = ["ps", "ux"]

    if all_users:
        cmd[1] = "aux"

    # pylint: disable=consider-using-with
    try:
        process_status = subprocess.Popen(cmd, stdout=subprocess.PIPE, encoding="utf8").communicate()[0]
    except TypeError:
        process_status = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
    # pylint: enable=consider-using-with

    if name in process_status:
        return True

    return False


def expandvars2(path: str) -> str:
    """
    Replace shell variables in the given path with their corresponding
    environment variable values.

    This function expands shell-style variable references (e.g., $VAR)
    in the input path using the current environment variables. It also
    ensures that any escaped dollar signs (e.g., \\$) are not expanded.

    Args:
        path: The input path containing shell variable references to be expanded.

    Returns:
        The path with shell variables replaced by their corresponding values
            from the environment, with unescaped variables expanded.
    """
    return re.sub(r"(?<!\\)\$[A-Za-z_][A-Za-z0-9_]*", "", os.path.expandvars(path))


def regex_list_filter(regex: str, list_to_filter: List[str], match: bool = True) -> List[str]:
    """
    Apply a regex filter to a list.

    This function filters a given list based on a specified regular expression.
    Depending on the `match` parameter, it can either match the entire string
    or search for the regex pattern within the strings of the list.

    Args:
        regex: The regular expression to use for filtering the list.
        list_to_filter: The list of strings to be filtered based on the regex.
        match: If True, uses re.match to filter items that match the regex from
            the start. If False, uses re.search to filter items that contain the
            regex pattern.

    Returns:
        A new list containing the filtered items that match the regex.
    """
    r = re.compile(regex)  # pylint: disable=C0103
    if match:
        return list(filter(r.match, list_to_filter))
    return list(filter(r.search, list_to_filter))


def apply_list_of_regex(
    regex_list: List[str], list_to_filter: List[str], result_list: List[str], match: bool = False, display_warning: bool = True
):
    """
    Apply a list of regex patterns to a list and accumulate the results.

    This function takes each regex from the provided list of regex patterns
    and applies it to the specified list. The results of each successful
    match or search are appended to a result list. Optionally, it can display
    a warning if a regex does not match any item in the list.

    Args:
        regex_list: A list of regular expressions to apply to the `list_to_filter`.
        list_to_filter: The list of strings that the regex patterns will be applied to.
        result_list: The list where results of the regex filters will be appended.
        match: If True, uses re.match for applying the regex. If False, uses re.search.
        display_warning: If True, displays a warning message when no matches are
            found for a regex.

    Side Effect:
        This function modifies the `result_list` in place.
    """
    for regex in regex_list:
        filter_results = set(regex_list_filter(regex, list_to_filter, match))

        if not filter_results:
            if display_warning:
                LOG.warning(f"No regex match for {regex}.")
        else:
            result_list += filter_results


def load_yaml(filepath: str) -> Dict:
    """
    Safely read a YAML file and return its contents.

    Args:
        filepath: The file path to the YAML file to be read.

    Returns:
        A dict representing the contents of the YAML file.
    """
    with open(filepath, "r") as _file:
        return yaml.safe_load(_file)


def get_yaml_var(entry: Dict[str, Any], var: str, default: Any) -> Any:
    """
    Retrieve the value associated with a specified key from a YAML dictionary.

    This function attempts to return the value of `var` from the provided `entry`
    dictionary. If the key does not exist, it will try to access it as an attribute
    of the entry object. If neither is found, the function returns the specified
    `default` value.

    Args:
        entry: A dictionary representing the contents of a YAML file.
        var: The key or attribute name to retrieve from the entry.
        default: The default value to return if the key or attribute is not found.

    Returns:
        The value associated with `var` in the entry, or `default` if not found.
    """
    try:
        return entry[var]
    except (TypeError, KeyError):
        try:
            return getattr(entry, var)
        except AttributeError:
            return default


def load_array_file(filename: str, ndmin: int = 2) -> np.ndarray:
    """
    Load an array from a file based on its extension.

    This function reads an array stored in the specified `filename`.
    It supports three file types based on their extensions:

    - `.npy` for NumPy binary files
    - `.csv` for comma-separated values
    - `.tab` for whitespace (or tab) separated values

    The function ensures that the loaded array has at least `ndmin` dimensions.
    If the array is in binary format, it checks the dimensions without altering the data.

    Args:
        filename: The path to the file to load.
        ndmin: The minimum number of dimensions the array should have.

    Returns:
        The loaded array.

    Raises:
        TypeError: If the file extension is not one of the supported types
            (`.npy`, `.csv`, `.tab`).
    """
    protocol = determine_protocol(filename)

    # Don't change binary-stored numpy arrays; just check dimensions
    if protocol == "npy":
        array = np.load(filename, allow_pickle=True)
        if array.ndim < ndmin:
            LOG.error(
                f"Array in {filename} has fewer than the required \
                       minimum dimensions ({array.ndim} < {ndmin})!"
            )
    # Make sure text files load as strings with minimum number of dimensions
    elif protocol == "csv":
        array = np.loadtxt(filename, delimiter=",", ndmin=ndmin, dtype=str)
    elif protocol == "tab":
        array = np.loadtxt(filename, ndmin=ndmin, dtype=str)
    else:
        raise TypeError(
            f"{protocol} is not a valid array file extension.\
                         Choices: {ARRAY_FILE_FORMATS}"
        )

    return array


def determine_protocol(fname: str) -> str:
    """
    Determine the file protocol based on the file name extension.

    Args:
        fname: The name of the file whose protocol is to be determined.

    Returns:
        The protocol corresponding to the file extension (e.g., 'hdf5').

    Raises:
        ValueError: If the provided file name does not have a valid extension.
    """
    _, ext = os.path.splitext(fname)
    if ext.startswith("."):
        protocol = ext.lower().strip(".")
    else:
        raise ValueError(f"{fname} needs an ext (eg .hdf5) to determine protocol!")
    # Map .h5 to .hdf5
    if protocol == "h5":
        protocol = "hdf5"
    return protocol


def verify_filepath(filepath: str) -> str:
    """
    Verify that the given file path is valid and return its absolute form.

    This function checks if the specified `filepath` points to an existing file.
    It expands any user directory shortcuts (e.g., `~`) and environment variables
    in the provided path before verifying its existence. If the file does not exist,
    a ValueError is raised.

    Args:
        filepath: The path of the file to verify.

    Returns:
        The verified absolute file path with expanded environment variables.

    Raises:
        ValueError: If the provided file path does not point to a valid file.
    """
    filepath = os.path.abspath(os.path.expandvars(os.path.expanduser(filepath)))
    if not os.path.isfile(filepath):
        raise ValueError(f"'{filepath}' is not a valid filepath")
    return filepath


def verify_dirpath(dirpath: str) -> str:
    """
    Verify that the given directory path is valid and return its absolute form.

    This function checks if the specified `dirpath` points to an existing directory.
    It expands any user directory shortcuts (e.g., `~`) and environment variables
    in the provided path before verifying its existence. If the directory does not exist,
    a ValueError is raised.

    Args:
        dirpath: The path of the directory to verify.

    Returns:
        The verified absolute directory path with expanded environment variables.

    Raises:
        ValueError: If the provided directory path does not point to a valid directory.
    """
    dirpath: str = os.path.abspath(os.path.expandvars(os.path.expanduser(dirpath)))
    if not os.path.isdir(dirpath):
        raise ValueError(f"'{dirpath}' is not a valid directory path")
    return dirpath


@contextmanager
def cd(path: str) -> Generator[None, None, None]:  # pylint: disable=C0103
    """
    Context manager for changing the current working directory.

    This context manager changes the current working directory to the specified `path`
    while executing the block of code within the context. Once the block is exited,
    it restores the original working directory.

    Args:
        path: The path to the directory to change to.

    Yields:
        Control is yielded back to the block of code within the context.
    """
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def pickle_data(filepath: str, content: Any):
    """
    Dump content to a pickle file.

    This function serializes the given `content` and writes it to a specified file
    in pickle format. The file is opened in write mode, which will overwrite any
    existing content in the file.

    Args:
        filepath: The path to the file where the content will be saved.
        content: The data to be serialized and saved to the pickle file.
    """
    with open(filepath, "w") as f:  # pylint: disable=C0103
        pickle.dump(content, f)


def get_source_root(filepath: str) -> str:
    """
    Find the absolute project path given a file path from within the project.

    This function determines the root directory of a project by analyzing the given
    file path. It works by traversing the directory structure upwards until it
    encounters a directory name that is not an integer, which is assumed to be the
    project root.

    Args:
        filepath: The file path from within the project for which to find the root.

    Returns:
        The absolute path to the root directory of the project. Returns None if
            the path corresponds to the root directory itself.
    """
    filepath = os.path.abspath(filepath)
    sep = os.path.sep
    if filepath == sep:
        return None

    parent = os.path.dirname(filepath)
    # Walk backwards testing for integers.
    break_point = parent.split(sep)[-1]  # Initial value for lgtm.com
    for _, _dir in enumerate(parent.split(sep)[::-1]):
        try:
            int(_dir)
        except ValueError:
            break_point = _dir
            break

    root, _ = parent.split(break_point)
    root = os.path.split(root)[0]
    return root


def ensure_directory_exists(**kwargs: Dict[Any, Any]) -> bool:
    """
    Ensure that the directory for the specified aggregate file exists.

    This function checks if the directory for the given `aggregate_file` exists.
    If it does not exist, the function creates the necessary directories.

    Args:
        **kwargs: Keyword arguments that must include:\n
            - `aggregate_file` (str): The file path for which the directory needs to be ensured.

    Returns:
        True if the directory already existed. False otherwise.
    """
    aggregate_bundle = kwargs["aggregate_file"]
    dirname = os.path.dirname(aggregate_bundle)

    if not os.path.exists(dirname):
        LOG.info(f"making directories to {dirname}.")
        os.makedirs(dirname)
        return False
    return True


def nested_dict_to_namespaces(dic: Dict) -> SimpleNamespace:
    """
    Convert a nested dictionary into a nested SimpleNamespace structure.

    This function recursively transforms a dictionary (which may contain other
    dictionaries) into a structure of SimpleNamespace objects. Each key in the
    dictionary becomes an attribute of a SimpleNamespace, allowing for attribute-style
    access to the data.

    Args:
        dic: The nested dictionary to be converted.

    Returns:
        A SimpleNamespace object representing the nested structure of the input dictionary.

    Raises:
        TypeError: If the input is not a dictionary.
    """

    def recurse(dic):
        if not isinstance(dic, dict):
            return dic
        for key, val in list(dic.items()):
            dic[key] = recurse(val)
        return SimpleNamespace(**dic)

    if not isinstance(dic, dict):
        raise TypeError(f"{dic} is not a dict")

    new_dic = deepcopy(dic)
    return recurse(new_dic)


def nested_namespace_to_dicts(namespaces: SimpleNamespace) -> Dict:
    """
    Convert a nested SimpleNamespace structure into a nested dictionary.

    This function recursively transforms a SimpleNamespace (which may contain
    other SimpleNamespaces) into a dictionary structure. Each attribute of the
    SimpleNamespace becomes a key in the resulting dictionary.

    Args:
        namespaces: The nested SimpleNamespace to be converted.

    Returns:
        A dictionary representing the nested structure of the input
            SimpleNamespace.

    Raises:
        TypeError: If the input is not a SimpleNamespace.
    """

    def recurse(namespaces):
        if not isinstance(namespaces, SimpleNamespace):
            return namespaces
        for key, val in list(namespaces.__dict__.items()):
            setattr(namespaces, key, recurse(val))
        return namespaces.__dict__

    if not isinstance(namespaces, SimpleNamespace):
        raise TypeError(f"{namespaces} is not a SimpleNamespace")

    new_ns = deepcopy(namespaces)
    return recurse(new_ns)


def get_flux_version(flux_path: str, no_errors: bool = False) -> str:
    """
    Retrieve the version of Flux as a string.

    This function executes the Flux binary located at `flux_path` with the
    "version" command and parses the output to return the version number.
    If the command fails or the Flux binary cannot be found, it can either
    raise an error or return a default version based on the `no_errors` flag.

    Args:
        flux_path: The full path to the Flux binary.
        no_errors: A flag to suppress error messages and exceptions. If set to
            True, errors will be logged but not raised.

    Returns:
        The version of Flux as a string.

    Raises:
        FileNotFoundError: If the Flux binary cannot be found and `no_errors`
            is set to False.
        ValueError: If the version cannot be determined from the output and
            `no_errors` is set to False.
    """
    cmd = [flux_path, "version"]

    process = None

    try:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, encoding="utf8").communicate()  # pylint: disable=R1732
    except FileNotFoundError as e:  # pylint: disable=C0103
        if not no_errors:
            LOG.error(f"The flux path {flux_path} canot be found")
            LOG.error("Suppress this error with no_errors=True")
            raise e

    try:
        flux_ver = re.search(r"\s*([\d.]+)", process[0]).group(1)
    except (ValueError, TypeError) as e:  # pylint: disable=C0103
        if not no_errors:
            LOG.error("The flux version cannot be determined")
            LOG.error("Suppress this error with no_errors=True")
            raise e
        flux_ver = DEFAULT_FLUX_VERSION
        LOG.warning(f"Using syntax for default version: {flux_ver}")

    return flux_ver


def get_flux_cmd(flux_path: str, no_errors: bool = False) -> str:
    """
    Generate the Flux run command based on the installed version.

    This function determines the appropriate Flux command to use for
    running jobs, depending on the version of Flux installed at the
    specified `flux_path`. It defaults to "flux run" for versions
    greater than or equal to 0.48.x. For older versions, it adjusts
    the command accordingly.

    Args:
        flux_path: The full path to the Flux binary.
        no_errors: A flag to suppress error messages and exceptions
            if set to True.

    Returns:
        The appropriate Flux run command as a string.
    """
    # The default is for flux version >= 0.48.x
    # this may change in the future.
    flux_cmd = "flux run"

    flux_ver = get_flux_version(flux_path, no_errors=no_errors)

    vers = [int(n) for n in flux_ver.split(".")]
    if vers[0] == 0 and vers[1] < 48:
        flux_cmd = "flux mini run"

    if vers[0] == 0 and vers[1] < 13:
        flux_cmd = "flux wreckrun"

    return flux_cmd


def get_flux_alloc(flux_path: str, no_errors: bool = False) -> str:
    """
    Generate the `flux alloc` command based on the installed version.

    This function constructs the appropriate command for allocating
    resources with Flux, depending on the version of Flux installed
    at the specified `flux_path`. It defaults to "{flux_path} alloc"
    for versions greater than or equal to 0.48.x. For older versions,
    it adjusts the command accordingly.

    Args:
        flux_path: The full path to the Flux binary.
        no_errors: A flag to suppress error messages and exceptions
            if set to True.

    Returns:
        The appropriate Flux allocation command as a string.
    """
    # The default is for flux version >= 0.48.x
    # this may change in the future.
    flux_alloc = f"{flux_path} alloc"

    flux_ver = get_flux_version(flux_path, no_errors=no_errors)

    vers = [int(n) for n in flux_ver.split(".")]

    if vers[0] == 0 and vers[1] < 48:
        flux_alloc = f"{flux_path} mini alloc"

    return flux_alloc


def check_machines(machines: Union[str, List[str], Tuple[str]]) -> bool:
    """
    Check if the current machine is in the list of specified machines.

    This function determines whether the hostname of the current
    machine matches any entry in a provided list of machine names.
    It returns True if a match is found, otherwise it returns False.

    Args:
        machines: A single machine name or a list/tuple of machine
            names to compare with the current machine's hostname.

    Returns:
        True if the current machine's hostname matches any of the
            specified machines; False otherwise.
    """
    local_hostname = socket.gethostname()

    if not isinstance(machines, (list, tuple)):
        machines = [machines]

    for mach in machines:
        if mach in local_hostname:
            return True

    return False


def contains_token(string: str) -> bool:
    """
    Check if the given string contains a token of the form $(STR).

    This function uses a regular expression to search for tokens
    that match the pattern $(<word\\>), where <word\\> consists of
    alphanumeric characters and underscores. It returns True if
    such a token is found; otherwise, it returns False.

    Args:
        string: The input string to be checked for tokens.

    Returns:
        True if the input string contains a token of the form
            $(STR); False otherwise.
    """
    if re.search(r"\$\(\w+\)", string):
        return True
    return False


def contains_shell_ref(string: str) -> bool:
    """
    Check if the given string contains a shell variable reference.

    This function searches for shell variable references in the
    format of $<variable\\> or ${<variable\\>}, where <variable\\>
    consists of alphanumeric characters and underscores. It returns
    True if a match is found; otherwise, it returns False.

    Args:
        string: The input string to be checked for shell
            variable references.

    Returns:
        True if the input string contains a shell variable
            reference of the form $STR or ${STR}; False otherwise.
    """
    if re.search(r"\$\w+", string) or re.search(r"\$\{\w+\}", string):
        return True
    return False


def needs_merlin_expansion(cmd: str, restart_cmd: str, labels: List[str], include_sample_keywords: bool = True) -> bool:
    """
    Check if the provided command or restart command contains variables that require expansion.

    This function checks both the command (`cmd`) and the restart command (`restart_cmd`)
    for the presence of specified labels or sample keywords that indicate a need for variable
    expansion.

    Args:
        cmd: The command inside a study step to check for variable expansion.
        restart_cmd: The restart command inside a study step to check for variable expansion.
        labels: A list of labels to check for inside `cmd` and `restart_cmd`.
        include_sample_keywords: Flag to indicate whether to include default sample keywords
            in the label check.

    Returns:
        True if either `cmd` or `restart_cmd` contains any of the specified labels
            or default sample keywords, indicating a need for expansion. False otherwise.
    """
    sample_keywords = ["MERLIN_SAMPLE_ID", "MERLIN_SAMPLE_PATH", "merlin_sample_id", "merlin_sample_path"]
    if include_sample_keywords:
        labels += sample_keywords

    for label in labels:
        if f"$({label})" in cmd:
            return True
        # The restart may need expansion while the cmd does not.
        if restart_cmd and f"$({label})" in restart_cmd:
            return True

    # If we got through all the labels and no expansion was needed then these commands don't need expansion
    return False


def dict_deep_merge(dict_a: Dict, dict_b: Dict, path: str = None, conflict_handler: Callable = None):
    """
    Recursively merges `dict_b` into `dict_a`, performing a deep merge.

    This function combines two dictionaries by recursively merging
    the contents of `dict_b` into `dict_a`. Unlike Python's built-in
    dictionary merge, this function performs a deep merge, meaning
    it will merge nested dictionaries instead of just updating top-level keys.
    Existing keys in `dict_a` will not be updated unless a conflict handler
    is provided to resolve key conflicts.

    Credit to [this stack overflow post](https://stackoverflow.com/a/7205107).

    Args:
        dict_a: The dictionary that will be merged into.
        dict_b: The dictionary to merge into `dict_a`.
        path: The current path in the dictionary tree. This is used for logging
            purposes during recursion.
        conflict_handler: A function to handle conflicts when both dictionaries
            have the same key with different values. The function should return
            the value to be used in the merged dictionary. If not provided, a
            warning will be logged for conflicts.
    """

    # Check to make sure we have valid dict_a and dict_b input
    msgs = [
        f"{name} '{actual_dict}' is not a dict"
        for name, actual_dict in [("dict_a", dict_a), ("dict_b", dict_b)]
        if not isinstance(actual_dict, dict)
    ]
    if len(msgs) > 0:
        LOG.warning(f"Problem with dict_deep_merge: {', '.join(msgs)}. Ignoring this merge call.")
        return

    if path is None:
        path = []
    for key in dict_b:
        if key in dict_a:
            if isinstance(dict_a[key], dict) and isinstance(dict_b[key], dict):
                dict_deep_merge(dict_a[key], dict_b[key], path=path + [str(key)], conflict_handler=conflict_handler)
            elif isinstance(dict_a[key], list) and isinstance(dict_a[key], list):
                dict_a[key] += dict_b[key]
            elif dict_a[key] == dict_b[key]:
                pass  # same leaf value
            else:
                if conflict_handler is not None:
                    merged_val = conflict_handler(
                        dict_a_val=dict_a[key], dict_b_val=dict_b[key], key=key, path=path + [str(key)]
                    )
                    dict_a[key] = merged_val
                else:
                    # Want to just output a warning instead of raising an exception so that the workflow doesn't crash
                    LOG.warning(f"Conflict at {'.'.join(path + [str(key)])}. Ignoring the update to key '{key}'.")
        else:
            dict_a[key] = dict_b[key]


def find_vlaunch_var(vlaunch_var: str, step_cmd: str, accept_no_matches: bool = False) -> str:
    """
    Find and return the specified VLAUNCHER variable from the step command.

    This function searches for a variable defined in the VLAUNCHER context
    within the provided step command string. It looks for the variable in
    the format `MERLIN_<vlaunch_var>=<value>`. If the variable is found,
    it returns the variable in a format suitable for use in a command string.
    If the variable is not found, the behavior depends on the `accept_no_matches` flag.

    Args:
        vlaunch_var: The name of the VLAUNCHER variable (without the prefix 'MERLIN_').
        step_cmd: The command string of a step where the variable may be defined.
        accept_no_matches: If True, returns None if the variable is not found.
            If False, raises a ValueError. Defaults to False.

    Returns:
        The variable in the format '${MERLIN_<vlaunch_var\\>}' if found, otherwise None
            (if `accept_no_matches` is True) or raises a ValueError (if False).

    Raises:
        ValueError: If the variable is not found and `accept_no_matches` is False.
    """
    matches = list(re.findall(rf"^(?!#).*MERLIN_{vlaunch_var}=\d+", step_cmd, re.MULTILINE))

    if matches:
        return f"${{MERLIN_{vlaunch_var}}}"

    if accept_no_matches:
        return None
    raise ValueError(f"VLAUNCHER used but could not find MERLIN_{vlaunch_var} in the step.")


# Time utilities
def convert_to_timedelta(timestr: Union[str, int]) -> timedelta:
    """
    Convert a time string or integer to a timedelta object.

    The function takes a time string formatted as
    '[days]:[hours]:[minutes]:seconds', where days, hours, and minutes
    are optional. If an integer is provided, it is interpreted as the
    total number of seconds.

    Args:
        timestr: The time string in the specified format or an integer
            representing seconds.

    Returns:
        A timedelta object representing the duration specified by the input
            string or integer.

    Raises:
        ValueError: If the input string does not conform to the expected
            format or contains more than four time fields.
    """
    # make sure it's a string in case we get an int
    timestr = str(timestr)

    # remove time unit characters (if any exist)
    time_unit_chars = r"[dhms]"
    timestr = re.sub(time_unit_chars, "", timestr)

    nfields = len(timestr.split(":"))
    if nfields > 4:
        raise ValueError(f"Cannot convert {timestr} to a timedelta. Valid format: days:hours:minutes:seconds.")
    _, d, h, m, s = (":0" * 10 + timestr).rsplit(":", 4)  # pylint: disable=C0103
    tdelta = timedelta(days=int(d), hours=int(h), minutes=int(m), seconds=int(s))
    return tdelta


def _repr_timedelta_HMS(time_delta: timedelta) -> str:  # pylint: disable=C0103
    """
    Represent a timedelta object as a string in 'HH:MM:SS' format.

    This function converts a given timedelta object into a string that
    represents the duration in hours, minutes, and seconds. The output
    is formatted as 'HH:MM:SS', with leading zeros for single-digit
    hours, minutes, or seconds.

    Args:
        time_delta: The timedelta object to be converted.

    Returns:
        A string representation of the timedelta in the format 'HH:MM:SS'.
    """
    hours, remainder = divmod(time_delta.total_seconds(), 3600)
    minutes, seconds = divmod(remainder, 60)
    hours, minutes, seconds = int(hours), int(minutes), int(seconds)
    return f"{hours:02d}:{minutes:02d}:{seconds:02d}"


def _repr_timedelta_FSD(time_delta: timedelta) -> str:  # pylint: disable=C0103
    """
    Represent a timedelta as a Flux Standard Duration (FSD) string in seconds.

    The FSD format represents a duration as a floating-point number followed
    by a suffix indicating the time unit. This function simplifies the
    representation by using seconds and appending an 's' suffix.

    Args:
        time_delta: The timedelta object to be converted.

    Returns:
        A string representation of the timedelta in FSD format, expressed
            in seconds (e.g., '123.45s').
    """
    fsd = f"{time_delta.total_seconds()}s"
    return fsd


def repr_timedelta(time_delta: timedelta, method: str = "HMS") -> str:
    """
    Represent a timedelta object as a string using a specified format method.

    This function formats a given timedelta object according to the chosen
    method. The available methods are:

    - HMS: Represents the duration in 'hours:minutes:seconds' format.
    - FSD: Represents the duration in Flux Standard Duration (FSD),
      expressed as a floating-point number of seconds with an 's' suffix.

    Args:
        time_delta: The timedelta object to be formatted.
        method: The method to use for formatting. Must be either 'HMS' or 'FSD'.

    Returns:
        A string representation of the timedelta formatted according
            to the specified method.

    Raises:
        ValueError: If an invalid method is provided.
    """
    if method == "HMS":
        return _repr_timedelta_HMS(time_delta)
    if method == "FSD":
        return _repr_timedelta_FSD(time_delta)
    raise ValueError("Invalid method for formatting timedelta! Valid choices: HMS, FSD")


def convert_timestring(timestring: Union[str, int], format_method: str = "HMS") -> str:
    """
    Converts a timestring to a specified format.

    This function accepts a timestring in a specific format or an integer
    representing seconds, and converts it to a formatted string based on
    the chosen format method. The available format methods are:

    - HMS: Represents the duration in 'hours:minutes:seconds' format.
    - FSD: Represents the duration in Flux Standard Duration (FSD),
      expressed as a floating-point number of seconds with an 's' suffix.

    Args:
        timestring: A string representing time in the format
            '[days]:[hours]:[minutes]:seconds' (where days, hours, and
            minutes are optional), or an integer representing time in seconds.
        format_method: The method to use for formatting. Must be either
            'HMS' or 'FSD'.

    Returns:
        A string representation of the converted timestring formatted
            according to the specified method.
    """
    LOG.debug(f"Timestring is: {timestring}")
    tdelta = convert_to_timedelta(timestring)
    LOG.debug(f"Timedelta object is: {tdelta}")
    return repr_timedelta(tdelta, method=format_method)


def pretty_format_hms(timestring: str) -> str:
    """
    Format an HMS timestring to remove blank entries and add appropriate labels.

    This function takes a timestring in the 'HH:MM:SS' format and formats
    it by removing any components that are zero and appending the relevant
    labels (days, hours, minutes, seconds). The output is a cleaner string
    representation of the time.

    Args:
        timestring: A timestring formatted as 'DD:HH:MM:SS'. Each component
            represents days, hours, minutes, and seconds, respectively.
            Only the last four components are relevant and may include
            leading zeros.

    Returns:
        A formatted timestring with non-zero components labeled appropriately.

    Raises:
        ValueError: If the input timestring contains more than four components
            or is not in the expected format.

    Examples:
        ```python
        >>> pretty_format_hms("00:00:34:00")
        '34m'
        >>> pretty_format_hms("01:00:00:25")
        '01d:25s'
        >>> pretty_format_hms("00:19:44:28")
        '19h:44m:28s'
        >>> pretty_format_hms("00:00:00:00")
        '00s'
        ```
    """
    # Create labels and split the timestring
    labels = ["d", "h", "m", "s"]
    parsed_ts = timestring.split(":")
    if len(parsed_ts) > 4:
        raise ValueError("The timestring to label must be in the format DD:HH:MM:SS")

    # Label each integer with its corresponding unit
    labeled_time_list = []
    for i in range(1, len(parsed_ts) + 1):
        if parsed_ts[-i] != "00":
            labeled_time_list.append(parsed_ts[-i] + labels[-i])

    # Join the labeled time list into a string.
    if len(labeled_time_list) == 0:
        labeled_time_list.append("00s")
    labeled_time_list.reverse()
    labeled_time_string = ":".join(labeled_time_list)

    return labeled_time_string


def ws_time_to_dt(ws_time: str) -> datetime:
    """
    Convert a workspace timestring to a datetime object.

    This function takes a workspace timestring formatted as 'YYYYMMDD-HHMMSS'
    and converts it into a corresponding datetime object. The input string
    must adhere to the specified format to ensure accurate conversion.

    Args:
        ws_time: A workspace timestring in the format 'YYYYMMDD-HHMMSS', where:\n
            - YYYY is the four-digit year,
            - MM is the two-digit month (01 to 12),
            - DD is the two-digit day (01 to 31),
            - HH is the two-digit hour (00 to 23),
            - MM is the two-digit minute (00 to 59),
            - SS is the two-digit second (00 to 59).

    Returns:
        A datetime object constructed from the provided workspace timestring.
    """
    year = int(ws_time[:4])
    month = int(ws_time[4:6])
    day = int(ws_time[6:8])
    hour = int(ws_time[9:11])
    minute = int(ws_time[11:13])
    second = int(ws_time[13:])
    return datetime(year, month, day, hour=hour, minute=minute, second=second)


def get_package_versions(package_list: List[str]) -> str:
    """
    Generate a formatted table of installed package versions and their locations.

    This function takes a list of package names and checks for their installed
    versions and locations. If a package is not installed, it indicates that
    the package is "Not installed". The output includes the Python version
    and its executable location at the top of the table.

    Args:
        package_list: A list of package names to check for installed versions.

    Returns:
        A formatted string representing a table of package names, their versions,
            and installation locations.
    """
    table = []
    for package in package_list:
        try:
            distribution = pkg_resources.get_distribution(package)
            version = distribution.version
            location = distribution.location
            table.append([package, version, location])
        except pkg_resources.DistributionNotFound:
            table.append([package, "Not installed", "N/A"])

    table.insert(0, ["python", sys.version.split()[0], sys.executable])
    table_str = tabulate(table, headers=["Package", "Version", "Location"], tablefmt="simple")
    return f"Python Packages\n\n{table_str}\n"



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Merlin CLI Package.

This package defines the core components and supporting utilities for the
Merlin command-line interface (CLI). It provides the entry point parser for
the `merlin` CLI tool, a suite of modular subcommands for workflow and
infrastructure management, and shared helper functions used across CLI
handlers.

Subpackages:
    commands: Contains all command implementations for the Merlin CLI, including
        workflow execution, monitoring, database interaction, and more.

Modules:
    argparse_main: Sets up the top-level argument parser and integrates all
        registered CLI subcommands into the `merlin` CLI interface.
    utils: Provides shared utility functions for parsing arguments, loading
        YAML specifications, and handling configuration logic across commands.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing database entities related to runs.

This module provides functionality for interacting with runs stored in a database,
including creating, retrieving, updating, and deleting runs. It defines the `RunEntity`
class, which extends the abstract base class [`DatabaseEntity`][db_scripts.entities.db_entity.DatabaseEntity],
to encapsulate run-specific operations and behaviors.
"""

import logging
import os
from typing import List

from merlin.backends.results_backend import ResultsBackend
from merlin.db_scripts.data_models import RunModel
from merlin.db_scripts.entities.db_entity import DatabaseEntity
from merlin.db_scripts.entities.mixins.queue_management import QueueManagementMixin
from merlin.exceptions import RunNotFoundError


LOG = logging.getLogger("merlin")


class RunEntity(DatabaseEntity[RunModel], QueueManagementMixin):
    """
    A class representing a run in the database.

    This class provides methods to interact with and manage a run's data, including
    retrieving information about the run, updating its state, and saving or deleting
    it from the database.

    Attributes:
        entity_info (db_scripts.data_models.RunModel): An instance of the `RunModel` class
            containing the run's metadata.
        backend (backends.results_backend.ResultsBackend): An instance of the `ResultsBackend`
            class used to interact with the database.
        run_complete (bool): A property to get or set the completion status of the run.

    Methods:
        __repr__:
            Provide a string representation of the `RunEntity` instance.

        __str__:
            Provide a human-readable string representation of the `RunEntity` instance.

        reload_data:
            Reload the latest data for this run from the database.

        get_id:
            Retrieve the ID of the run. _Implementation found in
                [`DatabaseEntity.get_id`][db_scripts.entities.db_entity.DatabaseEntity.get_id]._

        get_additional_data:
            Retrieve any additional data saved to this run. _Implementation found in
                [`DatabaseEntity.get_additional_data`][db_scripts.entities.db_entity.DatabaseEntity.get_additional_data]._

        get_metadata_file:
            Retrieve the path to the metadata file for this run.

        get_metadata_filepath:
            (classmethod) Retrieve the path to the metadata file for a given workspace.

        get_study_id:
            Retrieve the ID of the study associated with this run.

        get_workspace:
            Retrieve the path to the output workspace for this run.

        get_queues:
            Retrieve the task queues used for this run.

        get_workers:
            Retrieve the workers used for this run.

        add_worker:
            Add a worker to the list of workers used for this run.

        remove_worker:
            Remove a worker from the list of workers used for this run.

        get_parent:
            Retrieve the ID of the parent run that launched this run (if any).

        get_child:
            Retrieve the ID of the child run launched by this run (if any).

        save:
            Save the current state of the run to the database and dump its metadata.

        dump_metadata:
            Dump all metadata for this run to a JSON file.

        load:
            (classmethod) Load a `RunEntity` instance from the database by its ID or workspace.

        delete:
            (classmethod) Delete a run from the database by its ID or workspace.
    """

    def __init__(self, run_info: RunModel, backend: ResultsBackend):
        """
        Initialize a `RunEntity` instance.

        Args:
            run_info (db_scripts.data_models.RunModel): The data model containing
                information about the run.
            backend (backends.results_backend.ResultsBackend): The backend instance used to
                interact with the database.
        """
        super().__init__(run_info, backend)
        self._metadata_file = self.get_metadata_filepath(self.get_workspace())

    @classmethod
    def _get_entity_type(cls) -> str:
        return "run"

    def __repr__(self) -> str:
        """
        Provide a string representation of the `RunEntity` instance.

        Returns:
            A human-readable string representation of the `RunEntity` instance.
        """
        return (
            f"RunEntity("
            f"id={self.get_id()}, "
            f"study_id={self.get_study_id()}, "
            f"workspace={self.get_workspace()}, "
            f"queues={self.get_queues()}, "
            f"workers={self.get_workers()}, "
            f"parent={self.get_parent()}, "
            f"child={self.get_child()}, "
            f"run_complete={self.run_complete}, "
            f"additional_data={self.get_additional_data()}, "
            f"backend={self.backend.get_name()})"
        )

    def __str__(self) -> str:
        """
        Provide a string representation of the `RunEntity` instance.

        Returns:
            A human-readable string representation of the `RunEntity` instance.
        """
        from merlin.db_scripts.entities.study_entity import StudyEntity  # pylint: disable=import-outside-toplevel

        run_id = self.get_id()
        study = StudyEntity.load(self.get_study_id(), self.backend)
        study_str = f"  - ID: {study.get_id()}\n    Name: {study.get_name()}"
        return (
            f"Run with ID {run_id}\n"
            f"------------{'-' * len(run_id)}\n"
            f"Workspace: {self.get_workspace()}\n"
            f"Study:\n{study_str}\n"
            f"Queues: {self.get_queues()}\n"
            f"Workers: {self.get_workers()}\n"
            f"Parent: {self.get_parent()}\n"
            f"Child: {self.get_child()}\n"
            f"Run Complete: {self.run_complete}\n"
            f"Additional Data: {self.get_additional_data()}\n\n"
        )

    @property
    def run_complete(self) -> bool:
        """
        An attribute representing whether this run is complete.

        A "complete" study is a study that has executed all steps.

        Returns:
            True if the study is complete. False, otherwise.
        """
        self.reload_data()
        return self.entity_info.run_complete

    @run_complete.setter
    def run_complete(self, value: bool):
        """
        Update the run's completion status.

        Args:
            value: The completion status of the run.
        """
        self.entity_info.run_complete = value

    def get_metadata_file(self) -> str:
        """
        Get the path to the metadata file for this run.

        Returns:
            The path to the metadata file for this run
        """
        return self._metadata_file

    @classmethod
    def get_metadata_filepath(cls, workspace: str) -> str:
        """
        Get the path to the metadata file for a given workspace.
        This is needed for the [`load`][db_scripts.entities.run_entity.RunEntity.load]
        method when loading from workspace as it can't use the non-classmethod version
        of this method.

        Args:
            workspace: The workspace directory for the run.

        Returns:
            The path to the metadata file.
        """
        return os.path.join(workspace, "merlin_info", "run_metadata.json")

    def get_study_id(self) -> str:
        """
        Get the ID for the study associated with this run.

        Returns:
            The ID for the study associated with this run.
        """
        return self.entity_info.study_id

    def get_workspace(self) -> str:
        """
        Get the path to the output workspace for this run.

        Returns:
            A string representing the output workspace for this run.
        """
        return self.entity_info.workspace

    def get_workers(self) -> List[str]:
        """
        Get the logical workers that this run is using.

        Returns:
            A list of logical worker ids.
        """
        return self.entity_info.workers

    def add_worker(self, worker_id: str):
        """
        Add a new worker id to the list of workers.

        Args:
            worker_id: The id of the worker to add.
        """
        self.entity_info.workers.append(worker_id)
        self.save()

    def remove_worker(self, worker_id: str):
        """
        Remove a worker id from the list of workers associated with this run.

        Does *not* delete a [`LogicalWorkerEntity`][db_scripts.entities.logical_worker_entity.LogicalWorkerEntity]
        from the database. This will only remove the worker's id from the list in this run.

        Args:
            worker_id: The ID of the worker to remove.
        """
        self.reload_data()
        self.entity_info.workers.remove(worker_id)
        self.save()

    def get_parent(self) -> str:
        """
        Get the ID of the run that launched this run (if any).

        This will only be set for iterative workflows with greater than 1 iteration.

        Returns:
            The ID of the run that launched this run.
        """
        self.reload_data()
        return self.entity_info.parent

    def get_child(self) -> str:
        """
        Get the ID of the run that was launched by this run (if any).

        This will only be set for iterative workflows with greater than 1 iteration.

        Returns:
            The ID of the run that was launched by this run.
        """
        self.reload_data()
        return self.entity_info.child

    def _post_save_hook(self):
        """
        Hook called after saving the run entity.
        For runs, we also need to dump metadata to a file.
        """
        self.dump_metadata()

    def dump_metadata(self):
        """
        Dump all of the metadata for this run to a json file.
        """
        self.entity_info.dump_to_json_file(self.get_metadata_file())

    @classmethod
    def load(cls, entity_identifier: str, backend: ResultsBackend) -> "RunEntity":
        """
        Load a run from the database by id or workspace.

        Args:
            entity_identifier: The ID or workspace of the run to load.
            backend: A [`ResultsBackend`][backends.results_backend.ResultsBackend] instance.

        Returns:
            A `RunEntity` instance.

        Raises:
            (exceptions.RunNotFoundError): If an entry for run was not found in the database.
        """
        if os.path.isdir(entity_identifier) and os.path.exists(entity_identifier):  # Load from workspace
            LOG.debug("Retrieving run from workspace.")
            metadata_file = cls.get_metadata_filepath(entity_identifier)
            entity_info = RunModel.load_from_json_file(metadata_file)
        else:  # Load from ID
            LOG.debug("Retrieving run from backend.")
            entity_info = backend.retrieve(entity_identifier, "run")

        if not entity_info:
            raise RunNotFoundError(f"Run with ID or workspace {entity_identifier} not found in the database.")

        return cls(entity_info, backend)

    @classmethod
    def delete(cls, entity_identifier: str, backend: ResultsBackend):
        """
        Delete a run from the database by id or workpsace.

        Args:
            entity_identifier: The ID or workspace of the run to delete.
            backend: A [`ResultsBackend`][backends.results_backend.ResultsBackend] instance.
        """
        LOG.debug(f"Deleting run with id or workspace '{entity_identifier}' from the database...")
        self = cls.load(entity_identifier, backend)
        backend.delete(self.get_id(), "run")
        LOG.info(f"Run with id or workspace '{entity_identifier}' has been successfully deleted.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing entities with names.

This module provides a mixin class, `NameMixin`, designed to simplify the management
of names associated with an entity. The mixin can be used by any class that has the
necessary attributes to support name retrieval, specifically an `entity_info` object
containing a `name` attribute.
"""

# pylint: disable=too-few-public-methods


class NameMixin:
    """
    Mixin for entities that have a name.

    This mixin provides a method for retrieving the name associated with an entity.
    It assumes that the class using this mixin has an `entity_info` object containing
    a `name` attribute.

    Methods:
        get_name:
            Retrieve the name associated with the entity.
    """

    def get_name(self) -> str:
        """
        Get the name of this entity.

        Assumptions:
            - The class using this must have an `entity_info` object containing a `name` attribute

        Returns:
            The name of this entity.
        """
        return self.entity_info.name



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Functions for encrypting backend traffic.
"""
from typing import Any

import celery.backends.base
from celery.backends.base import Backend

from merlin.common.security import encrypt


encrypt.init_key()

# remember what the original encode / decode are so we can call it in our
# wrapper
old_encode = celery.backends.base.Backend.encode
old_decode = celery.backends.base.Backend.decode


def _encrypt_encode(*args, **kwargs) -> bytes:
    """
    Intercepts calls to the encode method of the Celery backend and encrypts
    the encoded data.

    This function wraps the original encode method, encrypting the result
    after encoding.

    Returns:
        The encrypted encoded data in bytes format.
    """
    return encrypt.encrypt(old_encode(*args, **kwargs))


def _decrypt_decode(self: Backend, payload: bytes) -> Any:
    """
    Intercepts calls to the decode method of the Celery backend and decrypts
    the payload before decoding.

    This function wraps the original decode method, decrypting the payload
    prior to decoding.

    Args:
        self: The instance of the backend from which the decode method is called.
        payload: The encrypted data to be decrypted.

    Returns:
        The decoded data after decryption. Can be any format.
    """
    return old_decode(self, encrypt.decrypt(payload))


def set_backend_funcs():
    """
    Sets the encode and decode methods of the Celery backend to custom
    implementations that handle encryption and decryption.

    This function replaces the default encode and decode methods with
    `_encrypt_encode` and `_decrypt_decode`, respectively, ensuring that
    all data processed by the Celery backend is encrypted and decrypted
    appropriately.
    """
    celery.backends.base.Backend.encode = _encrypt_encode
    celery.backends.base.Backend.decode = _decrypt_decode



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module provides functionality for managing and configuring connection strings
and SSL settings for various results backends, including MySQL, Redis, Rediss, and SQLite.
The module relies on the application's configuration file (`app.yaml`) to determine backend
settings and certificate paths.
"""
from __future__ import print_function

import logging
import os
from typing import Dict
from urllib.parse import quote

from merlin.config.config_filepaths import MERLIN_HOME
from merlin.config.configfile import CONFIG, get_ssl_entries


LOG = logging.getLogger(__name__)

BACKENDS = ["sqlite", "mysql", "redis", "rediss", "none"]


# Default files needed for the package to connect to the Rabbit instance.
MYSQL_CONFIG_FILENAMES = {
    "ssl_cert": "rabbit-client-cert.pem",
    "ssl_ca": "mysql-ca-cert.pem",
    "ssl_key": "rabbit-client-key.pem",
    "password": "mysql.password",
}


# fmt: off
MYSQL_CONNECTION_STRING = (
    "db+mysql+mysqldb://{user}:{password}@{server}/mlsi"
    "?ssl_ca={ssl_ca}"
    "&ssl_cert={ssl_cert}"
    "&ssl_key={ssl_key}"
)
# fmt: on

SQLITE_CONNECTION_STRING = os.path.join(MERLIN_HOME, "merlin.db")


def get_backend_password(password_file: str, certs_path: str = None) -> str:
    """
    Retrieves the backend password from a specified file or returns the provided password value.

    This function attempts to locate the password file in several locations:

    1. The default Merlin directory (`~/.merlin`).
    2. The path specified by `password_file`.
    3. A directory specified by `certs_path` (if provided).

    If the password file is found, the password is read from the file. If the file cannot be
    found, the value of `password_file` is treated as the password itself and returned.

    Args:
        password_file (str): The file path or value for the password. If this is not a valid
            file path, it is treated as the password itself.
        certs_path (str, optional): An optional directory path where SSL certificates and
            password files may be located.

    Returns:
        The backend password, either retrieved from the file or the provided value.
    """
    password = None

    mer_pass = os.path.join(os.path.expanduser("~/.merlin"), password_file)
    password_file = os.path.expanduser(password_file)

    password_filepath = ""
    if os.path.exists(mer_pass):
        password_filepath = mer_pass
    elif os.path.exists(password_file):
        password_filepath = password_file
    elif certs_path:
        password_filepath = os.path.join(certs_path, password_file)

    if not os.path.exists(password_filepath):
        # The password was given instead of the filepath.
        password = password_file.strip()
    else:
        with open(password_filepath, "r") as f:  # pylint: disable=C0103
            line = f.readline().strip()
            password = quote(line, safe="")

    LOG.debug(
        "Results backend: certs_path was provided and used in password resolution."
        if certs_path
        else "Results backend: certs_path was not provided."
    )
    LOG.debug("Password resolution: using file." if password_filepath else "Password resolution: using direct value.")

    return password


# flake8 complains about cyclomatic complexity because of all the try-excepts,
# this isn't so complicated it can't be followed and tucking things in functions
# would make it less readable, so complexity evaluation is off
def get_redis(certs_path: str = None, include_password: bool = True, ssl: bool = False) -> str:  # noqa C901
    """
    Constructs and returns a Redis or Rediss connection URL based on the provided parameters and configuration.

    Args:
        certs_path (str, optional): The path to SSL certificates and password files.
        include_password (bool, optional): Whether to include the password in the connection URL.
            If True, the password will be included; otherwise, it will be masked.
        ssl (bool, optional): Flag indicating whether to use SSL for the connection (Rediss).
            If True, the connection URL will use the "rediss" protocol; otherwise, it will use "redis".

    Returns:
        A Redis or Rediss connection URL formatted based on the provided parameters and configuration.
    """
    server = CONFIG.results_backend.server
    password_file = ""

    urlbase = "rediss" if ssl else "redis"

    try:
        port = CONFIG.results_backend.port
    except (KeyError, AttributeError):
        port = 6379
        LOG.debug("Results backend: using default Redis port.")

    try:
        db_num = CONFIG.results_backend.db_num
    except (KeyError, AttributeError):
        db_num = 0
        LOG.debug("Results backend: using default Redis database number.")

    try:
        username = CONFIG.results_backend.username
    except (KeyError, AttributeError):
        username = ""

    try:
        password_file = CONFIG.results_backend.password
        try:
            password = get_backend_password(password_file, certs_path=certs_path)
        except IOError:
            password = CONFIG.results_backend.password

        if include_password:
            spass = f"{username}:{password}@"
        else:
            spass = f"{username}:******@"
    except (KeyError, AttributeError):
        spass = ""
        LOG.debug("Results backend: no Redis password configured in backend config.")

    LOG.debug(
        f"Results backend: {'password file specified in config' if password_file else 'no password file specified; using direct value'}."
    )
    LOG.debug(f"Results backend: certs_path was {'provided' if certs_path else 'not provided'}.")
    LOG.debug(f"Results backend: Redis server address {'configured' if server else 'not found in config'}.")

    return f"{urlbase}://{spass}{server}:{port}/{db_num}"


def get_mysql_config(certs_path: str, mysql_certs: Dict) -> Dict:
    """
    Determines whether all required information for connecting to MySQL as the Celery
    results backend is available, and returns the MySQL SSL configuration or certificate paths.

    Args:
        certs_path (str): The path to the directory containing SSL certificates and password files.
        mysql_certs (Dict): A dictionary mapping certificate keys (e.g., 'cert', 'key', 'ca')
            to their expected filenames.

    Returns:
        A dictionary containing the paths to the required MySQL certificates if they exist.
    """
    mysql_ssl = get_ssl_config(celery_check=False)
    if mysql_ssl:
        return mysql_ssl

    if not os.path.exists(certs_path):
        return False
    files = os.listdir(certs_path)

    certs = {}
    for key, filename in mysql_certs.items():
        for f in files:  # pylint: disable=C0103
            if not f == filename:
                continue

            certs[key] = os.path.join(certs_path, f)

    return certs


def get_mysql(certs_path: str = None, mysql_certs: Dict = None, include_password: bool = True) -> str:
    """
    Constructs and returns a formatted MySQL connection string based on the provided parameters
    and application configuration.

    Args:
        certs_path (str, optional): The path to the directory containing SSL certificates and password files.
        mysql_certs (dict, optional): A dictionary mapping MySQL certificate keys (e.g., 'ssl_key', 'ssl_cert', 'ssl_ca')
            to their expected filenames. If this is None, it uses the default `MYSQL_CONFIG_FILENAMES`.
        include_password (bool, optional): Whether to include the password in the connection string.
            If True, the password will be included; otherwise, it will be masked.

    Returns:
        A formatted MySQL connection string.

    Raises:
        TypeError: \n
            - If the `server` configuration is missing or invalid.
            - If the MySQL connection information cannot be set due to missing certificates or configuration.
    """
    dbname = CONFIG.results_backend.dbname
    password_file = CONFIG.results_backend.password
    server = CONFIG.results_backend.server

    # Adding an initial start for printing configurations. This should
    # eventually be configured to use a logger. This logic should also
    # eventually be decoupled so we can print debug messages similar to our
    # Python debugging messages.
    LOG.debug(f"Results backend: database name is {'configured' if dbname else 'missing'}.")
    LOG.debug(
        f"Results backend: password file {'specified in configuration' if password_file else 'not specified in configuration; using direct value'}."
    )
    LOG.debug(f"Results backend: server address is {'configured' if server else 'missing'}.")
    LOG.debug(f"Results backend: certs_path was {'provided' if certs_path else 'not provided'}.")

    if not server:
        msg = f"Results backend: server {server} does not have a configuration"
        raise TypeError(msg)  # TypeError since server is None and not str

    password = get_backend_password(password_file, certs_path=certs_path)

    if mysql_certs is None:
        mysql_certs = MYSQL_CONFIG_FILENAMES

    mysql_config = get_mysql_config(certs_path, mysql_certs)

    if not mysql_config:
        msg = f"""The connection information for MySQL could not be set, cannot find:\n
        {mysql_certs}\ncheck the celery/certs path or set the ssl information in the app.yaml file."""
        raise TypeError(msg)  # TypeError since mysql_config is None when it shouldn't be

    mysql_config["user"] = CONFIG.results_backend.username
    if include_password:
        mysql_config["password"] = password
    else:
        mysql_config["password"] = "******"
    mysql_config["server"] = server

    # Ensure the ssl_key, ssl_ca, and ssl_cert keys are all set
    if mysql_certs == MYSQL_CONFIG_FILENAMES:
        for key, cert_file in mysql_certs.items():
            if key not in mysql_config:
                mysql_config[key] = os.path.join(certs_path, cert_file)

    return MYSQL_CONNECTION_STRING.format(**mysql_config)


def get_connection_string(include_password: bool = True) -> str:
    """
    Determines the appropriate results backend to use based on the package configuration
    and returns the corresponding connection string.

    If a URL is explicitly defined in the configuration (`CONFIG.results_backend.url`),
    it is returned as the connection string.

    Args:
        include_password (bool, optional): Whether to include the password in the connection string.
            If True, the password will be included; otherwise, it will be masked.

    Returns:
        The connection string for the configured results backend.

    Raises:
        ValueError: If the specified results backend in the configuration is not supported.
    """
    try:
        return CONFIG.results_backend.url
    except AttributeError:
        pass

    try:
        backend = CONFIG.results_backend.name.lower()
    except AttributeError:
        backend = ""

    if backend not in BACKENDS:
        msg = f"'{backend}' is not a supported results backend"
        raise ValueError(msg)

    try:
        certs_path = CONFIG.celery.certs
        certs_path = os.path.abspath(os.path.expanduser(certs_path))
    except AttributeError:
        certs_path = None

    return _resolve_backend_string(backend, certs_path, include_password)


def _resolve_backend_string(backend: str, certs_path: str, include_password: bool) -> str:
    """
    Resolves and returns the connection string for the specified results backend.

    Based on the backend type provided, this function delegates the connection string
    generation to the appropriate helper function or returns a predefined connection string.

    Args:
        backend (str): The name of the results backend (e.g., "mysql", "sqlite", "redis", "rediss").
        certs_path (str): The path to SSL certificates and password files, used for certain backends
            (e.g., MySQL and Redis).
        include_password (bool): Whether to include the password in the connection string.
            If True, the password will be included; otherwise, it will be masked.

    Returns:
        The connection string for the specified backend, or `None` if the backend is unsupported.
    """
    if "mysql" in backend:
        return get_mysql(certs_path=certs_path, include_password=include_password)

    if "sqlite" in backend:
        return SQLITE_CONNECTION_STRING

    if backend == "redis":
        return get_redis(certs_path=certs_path, include_password=include_password)

    if backend == "rediss":
        return get_redis(certs_path=certs_path, include_password=include_password, ssl=True)

    return None


def get_ssl_config(celery_check: bool = False) -> bool:
    """
    Retrieves the SSL configuration for the results backend based on the settings
    specified in the `app.yaml` configuration file.

    This function determines whether SSL should be enabled for the results backend
    and returns the appropriate configuration. It supports various backend types
    such as MySQL, Redis, and Rediss.

    Args:
        celery_check (bool, optional): If True, the function returns the SSL settings
            specifically for configuring Celery.

    Returns:
        The SSL configuration for the results backend. Returns `True` if SSL is enabled,
            `False` otherwise.
    """
    results_backend = ""
    try:
        results_backend = CONFIG.results_backend.url.split(":")[0]
    except AttributeError:
        # The results_backend may not have a url
        pass

    try:
        results_backend = CONFIG.results_backend.name.lower()
    except AttributeError:
        # The results_backend may not have a name
        pass

    if results_backend not in BACKENDS:
        return False

    try:
        certs_path = CONFIG.celery.certs
    except AttributeError:
        certs_path = None

    results_backend_ssl = get_ssl_entries("Results Backend", results_backend, CONFIG.results_backend, certs_path)

    if results_backend == "rediss":
        if not results_backend_ssl:
            results_backend_ssl = True
        return results_backend_ssl

    if results_backend and "mysql" in results_backend:
        if not celery_check:
            return results_backend_ssl

    return False



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module acts as an interface for users to interact with Merlin's
database.
"""

import logging
from argparse import Namespace
from typing import Any, List

from merlin.db_scripts.merlin_db import MerlinDatabase


LOG = logging.getLogger("merlin")


def database_info():
    """
    Print information about the database to the console.
    """
    merlin_db = MerlinDatabase()
    db_studies = merlin_db.get_all("study")
    db_runs = merlin_db.get_all("run")
    db_logical_workers = merlin_db.get_all("logical_worker")
    db_physical_workers = merlin_db.get_all("physical_worker")

    print("Merlin Database Information")
    print("---------------------------")
    print("General Information:")
    print(f"- Database Type: {merlin_db.get_db_type()}")
    print(f"- Database Version: {merlin_db.get_db_version()}")
    print(f"- Connection String: {merlin_db.get_connection_string()}")

    print()
    print("Studies:")
    print(f"- Total: {len(db_studies)}")
    # TODO add something about recent studies that looks like so:
    # - Recent Studies:
    #     1. Study ID: 123, Name: "Experiment A"
    #     2. Study ID: 124, Name: "Experiment B"
    #     3. Study ID: 125, Name: "Experiment C"
    #     (and 9 more studies)

    print()
    print("Runs:")
    print(f"- Total: {len(db_runs)}")
    # TODO add something about recent runs that looks like so:
    # - Recent Runs:
    #     1. Run ID: 456, Workspace: "/path/to/workspace"
    #     2. Run ID: 457, Workspace: "/path/to/workspace"
    #     3. Run ID: 458, Workspace: "/path/to/workspace"
    #     (and 42 more runs)

    print()
    print("Logical Workers:")
    print(f"- Total: {len(db_logical_workers)}")

    print()
    print("Physical Workers:")
    print(f"- Total: {len(db_physical_workers)}")

    print()


def database_get(args: Namespace):
    """
    Handles the delegation of get operations to Merlin's database.

    Args:
        args: Parsed CLI arguments from the user.
    """
    merlin_db = MerlinDatabase()

    def print_items(items: List[Any], empty_message: str):
        """
        Prints a list of items or logs a message if the list is empty.

        Args:
            items: List of items to print.
            empty_message: Message to log if the list is empty.
        """
        if items:
            for item in items:
                print(item)
        else:
            LOG.info(empty_message)

    def get_and_print(entity_type: str, identifiers: List[str]):
        """
        Get entities by type and identifiers, then print them.

        Args:
            entity_type: The entity type (study, run, logical_worker, etc.).
            identifiers: List of identifiers for fetching.
        """
        items = [merlin_db.get(entity_type, identifier) for identifier in identifiers]
        print_items(items, f"No {entity_type.replace('_', ' ')}s found for the given identifiers.")

    def get_all_and_print(entity_type: str):
        """
        Get all entities of a given type and print them.

        Args:
            entity_type: The entity type.
        """
        items = merlin_db.get_all(entity_type)
        entity_type_str = "studies" if entity_type == "study" else f"{entity_type.replace('_', ' ')}s"
        print_items(items, f"No {entity_type_str} found in the database.")

    operations = {
        "study": lambda: get_and_print("study", args.study),
        "run": lambda: get_and_print("run", args.run),
        "logical-worker": lambda: get_and_print("logical_worker", args.worker),
        "physical-worker": lambda: get_and_print("physical_worker", args.worker),
        "all-studies": lambda: get_all_and_print("study"),
        "all-runs": lambda: get_all_and_print("run"),
        "all-logical-workers": lambda: get_all_and_print("logical_worker"),
        "all-physical-workers": lambda: get_all_and_print("physical_worker"),
        "everything": lambda: print_items(merlin_db.get_everything(), "Nothing found in the database."),
    }

    operation = operations.get(args.get_type)
    if operation:
        operation()
    else:
        LOG.error("No valid get option provided.")


def database_delete(args: Namespace):
    """
    Handles the delegation of delete operations to Merlin's database.

    Args:
        args: Parsed CLI arguments from the user.
    """
    merlin_db = MerlinDatabase()

    def delete_entities(entity_type: str, identifiers: List[str], **kwargs):
        """
        Delete a list of entities by type and identifier.

        Args:
            entity_type: The entity type (study, run, etc.).
            identifiers: The identifiers of the entities to delete.
            kwargs: Additional keyword args for the delete call.
        """
        for identifier in identifiers:
            merlin_db.delete(entity_type, identifier, **kwargs)

    operations = {
        "study": lambda: delete_entities("study", args.study, remove_associated_runs=not args.keep_associated_runs),
        "run": lambda: delete_entities("run", args.run),
        "logical-worker": lambda: delete_entities("logical_worker", args.worker),
        "physical-worker": lambda: delete_entities("physical_worker", args.worker),
        "all-studies": lambda: merlin_db.delete_all("study", remove_associated_runs=not args.keep_associated_runs),
        "all-runs": lambda: merlin_db.delete_all("run"),
        "all-logical-workers": lambda: merlin_db.delete_all("logical_worker"),
        "all-physical-workers": lambda: merlin_db.delete_all("physical_worker"),
        "everything": lambda: merlin_db.delete_everything(force=args.force),
    }

    operation = operations.get(args.delete_type)
    if operation:
        operation()
    else:
        LOG.error("No valid delete option provided.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for inspecting task server queue statistics in Merlin workflows.

This module defines the `QueueInfoCommand` class, which implements the
`queue-info` subcommand for the Merlin CLI. The command enables users to query
detailed information about queues used in Merlin workflows, including the number
of tasks in each queue and the number of connected workers.
"""

# pylint: disable=duplicate-code

import logging
from argparse import ArgumentParser, Namespace

from tabulate import tabulate

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.cli.utils import get_merlin_spec_with_override
from merlin.router import dump_queue_info, query_queues


LOG = logging.getLogger("merlin")


class QueueInfoCommand(CommandEntryPoint):
    """
    Handles `queue-info` CLI command for querying info about the queues on the servers.

    Methods:
        add_parser: Adds the `queue-info` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `queue-info` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `queue-info` command parser will be added.
        """
        queue_info: ArgumentParser = subparsers.add_parser(
            "queue-info",
            help="List queue statistics (queue name, number of tasks in the queue, number of connected workers).",
        )
        queue_info.set_defaults(func=self.process_command)
        queue_info.add_argument(
            "--dump",
            type=str,
            help="Dump the queue information to a file. Provide the filename (must be .csv or .json)",
            default=None,
        )
        queue_info.add_argument(
            "--specific-queues", nargs="+", type=str, help="Display queue stats for specific queues you list here"
        )
        queue_info.add_argument(
            "--task_server",
            type=str,
            default="celery",
            help="Task server type. Default: %(default)s",
        )
        spec_group = queue_info.add_argument_group("specification options")
        spec_group.add_argument(
            "--spec",
            dest="specification",
            type=str,
            help="Path to a Merlin YAML spec file. \
                                This will only display information for queues defined in this spec file. \
                                This is the same behavior as the status command prior to Merlin version 1.11.0.",
        )
        spec_group.add_argument(
            "--steps",
            nargs="+",
            type=str,
            dest="steps",
            default=["all"],
            help="The specific steps in the YAML file you want to query the queues of. "
            "This option MUST be used with the --spec option",
        )
        spec_group.add_argument(
            "--vars",
            action="store",
            dest="variables",
            type=str,
            nargs="+",
            default=None,
            help="Specify desired Merlin variable values to override those found in the specification. Space-delimited. "
            "This option MUST be used with the --spec option. Example: '--vars LEARN=path/to/new_learn.py EPOCHS=3'",
        )

    def process_command(self, args: Namespace):
        """
        CLI command for finding all workers and their associated queues.

        This function processes the command-line arguments to retrieve and display
        information about the available workers and their queues within the task server.
        It validates the necessary parameters, handles potential file dumping, and
        formats the output for easy readability.

        Args:
            args: Parsed CLI arguments containing user inputs related to the query.

        Raises:
            ValueError:
                - If a specification is not provided when steps are specified and the
                steps do not include "all".
                - If variables are included without a corresponding specification.
                - If the specified dump filename does not end with '.json' or '.csv'.
        """
        print(banner_small)

        # Ensure a spec is provided if steps are provided
        if not args.specification:
            if "all" not in args.steps:
                raise ValueError("The --steps argument MUST be used with the --specification argument.")
            if args.variables:
                raise ValueError("The --vars argument MUST be used with the --specification argument.")

        # Ensure a supported file type is provided with the dump option
        if args.dump is not None:
            if not args.dump.endswith(".json") and not args.dump.endswith(".csv"):
                raise ValueError("Unsupported file type. Dump files must be either '.json' or '.csv'.")

        spec = None
        # Load the spec if necessary
        if args.specification:
            spec, _ = get_merlin_spec_with_override(args)

        # Obtain the queue information
        queue_information = query_queues(args.task_server, spec, args.steps, args.specific_queues)

        if queue_information:
            # Format the queue information so we can pass it to the tabulate library
            formatted_queue_info = [("Queue Name", "Task Count", "Worker Count")]
            for queue_name, queue_stats in queue_information.items():
                formatted_queue_info.append((queue_name, queue_stats["jobs"], queue_stats["consumers"]))

            # Print the queue information
            print()
            print(tabulate(formatted_queue_info, headers="firstrow"))
            print()

            # Dump queue information to an output file if necessary
            if args.dump:
                dump_queue_info(args.task_server, queue_information, args.dump)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `entities` package defines database entity classes used throughout Merlin for managing core
components such as studies, runs, and workers. These classes provide a structured interface for
interacting with persisted data, ensuring consistency and maintainability.

At the heart of this package is the abstract base class `DatabaseEntity`, which outlines the
standard methods that all database-backed entities must implement, including save, delete, and
reload operations.

Subpackages:
    - `mixins/`: Contains mixin classes for entities that help reduce shared code.

Modules:
    db_entity.py: Defines the abstract base class [`DatabaseEntity`][db_scripts.entities.db_entity.DatabaseEntity],
        which provides a common interface for all database entity classes.
    logical_worker_entity.py: Implements the
        [`LogicalWorkerEntity`][db_scripts.entities.logical_worker_entity.LogicalWorkerEntity]
        class, representing logical workers and their associated operations.
    physical_worker_entity.py: Defines the
        [`PhysicalWorkerEntity`][db_scripts.entities.physical_worker_entity.PhysicalWorkerEntity]
        class for managing physical workers stored in the database.
    run_entity.py: Implements the [`RunEntity`][db_scripts.entities.run_entity.RunEntity] class,
        which encapsulates database operations related to run records.
    study_entity.py: Defines the [`StudyEntity`][db_scripts.entities.study_entity.StudyEntity] class
        for handling study-related database interactions.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing logical worker entities.

This module defines the `LogicalWorkerManager` class, which provides high-level operations
for creating, retrieving, and deleting logical workers stored in the database. It extends
the generic [`EntityManager`][db_scripts.entity_managers.entity_manager.EntityManager] class
with logic specific to logical worker entities, such as ID resolution based on worker name
and queues.

The manager also integrates cleanup routines to maintain consistency across related entities,
e.g., by removing logical workers from associated runs before deletion.
"""

from __future__ import annotations

import logging
from typing import List

from merlin.db_scripts.data_models import LogicalWorkerModel
from merlin.db_scripts.entities.logical_worker_entity import LogicalWorkerEntity
from merlin.db_scripts.entity_managers.entity_manager import EntityManager
from merlin.exceptions import RunNotFoundError


LOG = logging.getLogger("merlin")

# Purposefully ignoring this pylint message as each entity will have different parameter requirements
# pylint: disable=arguments-differ,arguments-renamed


class LogicalWorkerManager(EntityManager[LogicalWorkerEntity, LogicalWorkerModel]):
    """
    Manager class for handling logical worker entities.

    This class provides methods to create, retrieve, and delete logical workers from the
    backend. It also handles internal ID resolution and cleanup of references to logical
    workers from related entities such as runs.

    Attributes:
        backend: The backend interface used for storing and retrieving logical workers.
        db: Reference to the main database interface, used for cross-entity operations
            such as detaching workers from runs.

    Methods:
        create: Create a logical worker with the given name and queue list.
        get: Retrieve a logical worker either by ID or by name and queues.
        get_all: Retrieve all logical workers in the system.
        delete: Delete a logical worker and remove it from all associated runs.
        delete_all: Delete all logical workers currently stored in the backend.
        set_db_reference: Set a reference to the main database object for accessing related entities.
    """

    def _resolve_worker_id(self, worker_id: str = None, worker_name: str = None, queues: List[str] = None) -> str:
        """
        Resolve the logical worker ID based on provided parameters.

        Either a `worker_id` must be provided, or both `worker_name` and `queues`.

        Args:
            worker_id (str, optional): The ID of the logical worker.
            worker_name (str, optional): The name of the logical worker.
            queues (List[str], optional): The queues the worker handles.

        Returns:
            The resolved logical worker ID.

        Raises:
            ValueError: If input arguments are invalid or insufficient.
        """
        # Same implementation as in the original class
        if worker_id is not None:
            if worker_name is not None or queues is not None:
                raise ValueError("Provide either `worker_id` or (`worker_name` and `queues`), but not both.")
            return worker_id
        if worker_name is None or queues is None:
            raise ValueError("You must provide either `worker_id` or both `worker_name` and `queues`.")

        return LogicalWorkerModel.generate_id(worker_name, queues)

    def create(self, name: str, queues: List[str]) -> LogicalWorkerEntity:
        """
        Create a new logical worker entity, or return it if it already exists.

        Args:
            name (str): The name of the logical worker.
            queues (List[str]): A list of queue names the worker is assigned to.

        Returns:
            The created or existing logical worker entity.
        """
        logical_worker_id = self._resolve_worker_id(worker_name=name, queues=queues)
        log_message_create = (
            f"Logical worker with name '{name}' and queues '{queues}' does not yet have "
            "an entry in the database. Creating one."
        )
        log_message_exists = f"Logical worker with name '{name}' and queues '{queues}' already has an entry in the database."
        return self._create_entity_if_not_exists(
            entity_class=LogicalWorkerEntity,
            model_class=LogicalWorkerModel,
            identifier=logical_worker_id,
            log_message_exists=log_message_exists,
            log_message_create=log_message_create,
            name=name,
            queues=queues,
        )

    def get(self, worker_id: str = None, worker_name: str = None, queues: List[str] = None) -> LogicalWorkerEntity:
        """
        Retrieve a logical worker entity by ID, or by name and queues.

        Args:
            worker_id (str, optional): The unique identifier of the logical worker.
            worker_name (str, optional): The name of the logical worker.
            queues (List[str], optional): The queues the worker handles.

        Returns:
            The retrieved logical worker entity.

        Raises:
            ValueError: If input arguments are invalid or insufficient.
            WorkerNotFoundError: If the specified worker does not exist.
        """
        worker_id = self._resolve_worker_id(worker_id=worker_id, worker_name=worker_name, queues=queues)
        return self._get_entity(LogicalWorkerEntity, worker_id)

    def get_all(self) -> List[LogicalWorkerEntity]:
        """
        Retrieve all logical worker entities from the backend.

        Returns:
            A list of all logical worker entities.
        """
        return self._get_all_entities(LogicalWorkerEntity, "logical_worker")

    def delete(self, worker_id: str = None, worker_name: str = None, queues: List[str] = None):
        """
        Delete a logical worker entity and clean up any references from associated runs.

        The method ensures the logical worker is removed from all runs that reference it
        before deleting it from the backend.

        Args:
            worker_id (str, optional): The ID of the logical worker.
            worker_name (str, optional): The name of the logical worker.
            queues (List[str], optional): The queues the worker handles.

        Raises:
            ValueError: If input arguments are invalid or insufficient.
        """
        logical_worker = self.get(worker_id=worker_id, worker_name=worker_name, queues=queues)

        def cleanup_logical_worker(worker):
            runs_using_worker = worker.get_runs()
            for run_id in runs_using_worker:
                try:
                    run = self.db.runs.get(run_id)
                    run.remove_worker(worker.get_id())
                except RunNotFoundError:
                    LOG.warning(f"Couldn't find run with id {run_id}. Continuing with logical worker delete.")

        self._delete_entity(LogicalWorkerEntity, logical_worker.get_id(), cleanup_fn=cleanup_logical_worker)

    def delete_all(self):
        """
        Delete all logical worker entities currently stored in the backend.

        Runs cleanup on each logical worker before deletion to remove dependencies.
        """
        self._delete_all_by_type(get_all_fn=self.get_all, delete_fn=self.delete, entity_name="logical workers")

    def set_db_reference(self, db: MerlinDatabase):  # noqa: F821  pylint: disable=undefined-variable
        """
        Set a reference to the main Merlin database object for cross-entity operations.

        This allows the manager to access other entity managers (e.g., runs) when
        performing operations like cleanup during deletions.

        Args:
            db (db_scripts.merlin_db.MerlinDatabase): The database object that provides
                access to related entity managers.
        """
        self.db = db



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""This module handles setting up the extensive logging system in Merlin."""

import logging
import sys

import coloredlogs


FORMATS = {
    "DEFAULT": "[%(asctime)s: %(levelname)s] %(message)s",
    "DEBUG": "[%(asctime)s: %(levelname)s] [%(module)s: %(lineno)d] %(message)s",
    "WORKER": "[%(asctime)s: %(levelname)s] [%(task_name)s(%(task_id)s)] %(message)s",
}


def setup_logging(logger: logging.Logger, log_level: str = "INFO", colors: bool = True):
    """
    Setup and configure Python logging.

    Args:
        logger: A logging.Logger object.
        log_level: Logger level.
        colors: If True use colored logs.
    """
    formatter = logging.Formatter()
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    logger.setLevel(log_level)
    logger.propagate = False

    if colors is True:
        coloredlogs.install(level=log_level, logger=logger, fmt=FORMATS["DEFAULT"])



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI command for managing Merlin configuration files.

This module defines the `ConfigCommand` class, which provides the CLI interface
to create, update, and switch between different Merlin configuration files.
These configurations control task server, broker, and backend settings used
by the Merlin workflow system.
"""

# pylint: disable=duplicate-code

import logging
import os
from argparse import SUPPRESS, ArgumentDefaultsHelpFormatter, ArgumentParser, ArgumentTypeError, Namespace

import yaml

from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.config.merlin_config_manager import MerlinConfigManager


LOG = logging.getLogger("merlin")


class ConfigCommand(CommandEntryPoint):
    """
    CLI command group for managing Merlin configuration files.

    This class adds the `config` command and its subcommands to the Merlin CLI,
    and dispatches the CLI input to Merlin's codebase.

    Attributes:
        default_config_file (str): The default path to the Merlin config file (`~/.merlin/app.yaml`).

    Methods:
        add_parser: Adds the `config` command and its subcommands to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    default_config_file = os.path.join(os.path.expanduser("~"), ".merlin", "app.yaml")

    def _add_create_subcommand(self, mconfig_subparsers: ArgumentParser):
        """
        Add the `create` subcommand to generate a new configuration file.

        Parameters:
            mconfig_subparsers (ArgumentParser): The subparsers object to add the subcommand to.
        """
        config_create_parser = mconfig_subparsers.add_parser("create", help="Create a new configuration file.")
        config_create_parser.add_argument(
            "--task-server",
            type=str,
            default="celery",
            help="Task server type for which to create the config. Default: %(default)s",
        )
        config_create_parser.add_argument(
            "-o",
            "--output-file",
            dest="config_file",
            type=str,
            default=self.default_config_file,
            help=f"Optional file name for your configuration. Default: {self.default_config_file}",
        )
        config_create_parser.add_argument(
            "--broker",
            type=str,
            default=None,
            help="Optional broker type, backend will be redis. Default: rabbitmq",
        )

    def _add_update_broker_subcommand(self, mconfig_subparsers: ArgumentParser):
        """
        Add the `update-broker` subcommand to modify broker-related settings.

        Parameters:
            mconfig_subparsers (ArgumentParser): The subparsers object to add the subcommand to.
        """
        config_broker_parser = mconfig_subparsers.add_parser("update-broker", help="Update broker settings in app.yaml")
        config_broker_parser.add_argument(
            "-t",
            "--type",
            required=True,
            choices=["redis", "rabbitmq"],
            help="Type of broker to configure (redis or rabbitmq).",
        )
        config_broker_parser.add_argument(
            "--cf",
            "--config-file",
            dest="config_file",
            default=self.default_config_file,
            help=f"The path to the config file that will be updated. Default: {self.default_config_file}",
        )
        config_broker_parser.add_argument("-u", "--username", help="Broker username (only for rabbitmq)")
        config_broker_parser.add_argument("--pf", "--password-file", dest="password_file", help="Path to password file")
        config_broker_parser.add_argument("-s", "--server", help="The URL of the server")
        config_broker_parser.add_argument("-p", "--port", type=int, help="Broker port")
        config_broker_parser.add_argument("-v", "--vhost", help="Broker vhost (only for rabbitmq)")
        config_broker_parser.add_argument("-c", "--cert-reqs", help="Broker cert requirements")
        config_broker_parser.add_argument("-d", "--db-num", type=int, help="Redis database number (only for redis).")

    def _add_update_backend_subcommand(self, mconfig_subparsers: ArgumentParser):
        """
        Add the `update-backend` subcommand to modify results backend settings.

        Parameters:
            mconfig_subparsers (ArgumentParser): The subparsers object to add the subcommand to.
        """
        config_backend_parser = mconfig_subparsers.add_parser(
            "update-backend", help="Update results backend settings in app.yaml"
        )
        config_backend_parser.add_argument(
            "-t",
            "--type",
            required=True,
            choices=["redis"],
            help="Type of results backend to configure.",
        )
        config_backend_parser.add_argument(
            "--cf",
            "--config-file",
            dest="config_file",
            default=self.default_config_file,
            help=f"The path to the config file that will be updated. Default: {self.default_config_file}",
        )
        config_backend_parser.add_argument("-u", "--username", help="Backend username")
        config_backend_parser.add_argument("--pf", "--password-file", dest="password_file", help="Path to password file")
        config_backend_parser.add_argument("-s", "--server", help="The URL of the server")
        config_backend_parser.add_argument("-p", "--port", help="Backend port")
        config_backend_parser.add_argument("-d", "--db-num", help="Backend database number")
        config_backend_parser.add_argument("-c", "--cert-reqs", help="Backend cert requirements")
        config_backend_parser.add_argument("-e", "--encryption-key", help="Path to encryption key file")

    def _add_use_subcommand(self, mconfig_subparsers: ArgumentParser):
        """
        Add the `use` subcommand to switch to a different Merlin config file.

        Parameters:
            mconfig_subparsers (ArgumentParser): The subparsers object to add the subcommand to.
        """
        config_use_parser = mconfig_subparsers.add_parser("use", help="Use a different configuration file.")
        config_use_parser.add_argument("config_file", type=str, help="The path to the new configuration file to use.")

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `config` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `config` command parser will be added.
        """
        mconfig: ArgumentParser = subparsers.add_parser(
            "config",
            help="Create a default merlin server config file in ~/.merlin",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        mconfig.set_defaults(func=self.process_command)
        # The below option makes it so the `config_path.txt` file is written to the test directory
        mconfig.add_argument(
            "-t",
            "--test",
            action="store_true",
            help=SUPPRESS,  # Hides from `--help`
        )
        mconfig_subparsers = mconfig.add_subparsers(dest="commands", help="Subcommands for 'config'")

        self._add_create_subcommand(mconfig_subparsers)
        self._add_update_broker_subcommand(mconfig_subparsers)
        self._add_update_backend_subcommand(mconfig_subparsers)
        self._add_use_subcommand(mconfig_subparsers)

    def process_command(self, args: Namespace):
        """
        CLI command to manage Merlin configuration files.

        This function handles various configuration-related operations based on
        the provided subcommand. It ensures that the specified configuration
        file has a valid YAML extension (i.e., `.yaml` or `.yml`).

        If no output file is explicitly provided, a default path is used.

        Args:
            args (Namespace): Parsed command-line arguments.
        """
        if args.commands != "create":  # Check that this is a valid yaml file
            try:
                with open(args.config_file, "r") as conf_file:
                    yaml.safe_load(conf_file)
            except FileNotFoundError as fnf_exc:
                raise ArgumentTypeError(f"The file '{args.config_file}' does not exist.") from fnf_exc
            except yaml.YAMLError as yaml_exc:
                raise ArgumentTypeError(f"The file '{args.config_file}' is not a valid YAML file.") from yaml_exc

        config_manager = MerlinConfigManager(args)

        if args.commands == "create":
            config_manager.create_template_config()
            config_manager.save_config_path()
        elif args.commands == "update-broker":
            config_manager.update_broker()
        elif args.commands == "update-backend":
            config_manager.update_backend()
        elif args.commands == "use":  # Config file path is updated in constructor of MerlinConfigManager
            config_manager.config_file = args.config_file
            config_manager.save_config_path()
